from .error import OutOfRangeError

# For most use cases in this module, numpy is indistinguishable
# from math, except it also works on numpy arrays
try:
    import numpy as mathlib

    use_numpy = True
except ImportError:
    import math as mathlib

    use_numpy = False

__all__ = ["to_latlon", "from_latlon"]

K0 = 0.9996

E = 0.00669438
E2 = E * E
E3 = E2 * E
E_P2 = E / (1.0 - E)

SQRT_E = mathlib.sqrt(1 - E)
_E = (1 - SQRT_E) / (1 + SQRT_E)
_E2 = _E * _E
_E3 = _E2 * _E
_E4 = _E3 * _E
_E5 = _E4 * _E

M1 = 1 - E / 4 - 3 * E2 / 64 - 5 * E3 / 256
M2 = 3 * E / 8 + 3 * E2 / 32 + 45 * E3 / 1024
M3 = 15 * E2 / 256 + 45 * E3 / 1024
M4 = 35 * E3 / 3072

P2 = 3. / 2 * _E - 27. / 32 * _E3 + 269. / 512 * _E5
P3 = 21. / 16 * _E2 - 55. / 32 * _E4
P4 = 151. / 96 * _E3 - 417. / 128 * _E5
P5 = 1097. / 512 * _E4

R = 6378137

ZONE_LETTERS = "CDEFGHJKLMNPQRSTUVWXX"


def in_bounds(x, lower, upper, upper_strict=False):
    if upper_strict and use_numpy:
        return lower <= mathlib.min(x) and mathlib.max(x) < upper
    elif upper_strict and not use_numpy:
        return lower <= x < upper
    elif use_numpy:
        return lower <= mathlib.min(x) and mathlib.max(x) <= upper
    return lower <= x <= upper


def check_valid_zone(zone_number, zone_letter):
    if not 1 <= zone_number <= 60:
        raise OutOfRangeError("zone number out of range (must be between 1 and 60)")

    if zone_letter:
        zone_letter = zone_letter.upper()

        if not "C" <= zone_letter <= "X" or zone_letter in ["I", "O"]:
            raise OutOfRangeError("zone letter out of range (must be between C and X)")


def mixed_signs(x):
    return use_numpy and mathlib.min(x) < 0 and mathlib.max(x) >= 0


def negative(x):
    if use_numpy:
        return mathlib.max(x) < 0
    return x < 0


def to_latlon(
    easting, northing, zone_number, zone_letter=None, northern=None, strict=True
):
    """This function convert an UTM coordinate into Latitude and Longitude

        Parameters
        ----------
        easting: int
            Easting value of UTM coordinate

        northing: int
            Northing value of UTM coordinate

        zone number: int
            Zone Number is represented with global map numbers of an UTM Zone
            Numbers Map. More information see utmzones [1]_

        zone_letter: str
            Zone Letter can be represented as string values. Where UTM Zone
            Designators can be accessed in [1]_

        northern: bool
            You can set True or False to set this parameter. Default is None


       .. _[1]: http://www.jaworski.ca/utmzones.htm

    """
    if not zone_letter and northern is None:
        raise ValueError("either zone_letter or northern needs to be set")

    elif zone_letter and northern is not None:
        raise ValueError("set either zone_letter or northern, but not both")

    if strict:
        if not in_bounds(easting, 100000, 1000000, upper_strict=True):
            raise OutOfRangeError(
                "easting out of range (must be between 100.000 m and 999.999 m)"
            )
        if not in_bounds(northing, 0, 10000000):
            raise OutOfRangeError(
                "northing out of range (must be between 0 m and 10.000.000 m)"
            )

    check_valid_zone(zone_number, zone_letter)

    if zone_letter:
        zone_letter = zone_letter.upper()
        northern = zone_letter >= "N"

    x = easting - 500000
    y = northing

    if not northern:
        y -= 10000000

    m = y / K0
    mu = m / (R * M1)

    p_rad = (
        mu +
        P2 * mathlib.sin(2 * mu) +
        P3 * mathlib.sin(4 * mu) +
        P4 * mathlib.sin(6 * mu) +
        P5 * mathlib.sin(8 * mu)
    )

    p_sin = mathlib.sin(p_rad)
    p_sin2 = p_sin * p_sin

    p_cos = mathlib.cos(p_rad)

    p_tan = p_sin / p_cos
    p_tan2 = p_tan * p_tan
    p_tan4 = p_tan2 * p_tan2

    ep_sin = 1 - E * p_sin2
    ep_sin_sqrt = mathlib.sqrt(1 - E * p_sin2)

    n = R / ep_sin_sqrt
    r = (1 - E) / ep_sin

    c = _E * p_cos ** 2
    c2 = c * c

    d = x / (n * K0)
    d2 = d * d
    d3 = d2 * d
    d4 = d3 * d
    d5 = d4 * d
    d6 = d5 * d

    latitude = (
        p_rad -
        (p_tan / r) *
        (d2 / 2 - d4 / 24 * (5 + 3 * p_tan2 + 10 * c - 4 * c2 - 9 * E_P2)) +
        d6 / 720 * (61 + 90 * p_tan2 + 298 * c + 45 * p_tan4 - 252 * E_P2 - 3 * c2)
    )

    longitude = (
        d -
        d3 / 6 * (1 + 2 * p_tan2 + c) +
        d5 / 120 * (5 - 2 * c + 28 * p_tan2 - 3 * c2 + 8 * E_P2 + 24 * p_tan4)
    ) / p_cos

    return (
        mathlib.degrees(latitude),
        mathlib.degrees(longitude) + zone_number_to_central_longitude(zone_number),
    )


def from_latlon(latitude, longitude, force_zone_number=None, force_zone_letter=None):
    """This function convert Latitude and Longitude to UTM coordinate

        Parameters
        ----------
        latitude: float
            Latitude between 80 deg S and 84 deg N, e.g. (-80.0 to 84.0)

        longitude: float
            Longitude between 180 deg W and 180 deg E, e.g. (-180.0 to 180.0).

        force_zone number: int
            Zone Number is represented with global map numbers of an UTM Zone
            Numbers Map. You may force conversion including one UTM Zone Number.
            More information see utmzones [1]_

       .. _[1]: http://www.jaworski.ca/utmzones.htm
    """
    if not in_bounds(latitude, -80.0, 84.0):
        raise OutOfRangeError(
            "latitude out of range (must be between 80 deg S and 84 deg N)"
        )
    if not in_bounds(longitude, -180.0, 180.0):
        raise OutOfRangeError(
            "longitude out of range (must be between 180 deg W and 180 deg E)"
        )
    if force_zone_number is not None:
        check_valid_zone(force_zone_number, force_zone_letter)

    lat_rad = mathlib.radians(latitude)
    lat_sin = mathlib.sin(lat_rad)
    lat_cos = mathlib.cos(lat_rad)

    lat_tan = lat_sin / lat_cos
    lat_tan2 = lat_tan * lat_tan
    lat_tan4 = lat_tan2 * lat_tan2

    if force_zone_number is None:
        zone_number = latlon_to_zone_number(latitude, longitude)
    else:
        zone_number = force_zone_number

    if force_zone_letter is None:
        zone_letter = latitude_to_zone_letter(latitude)
    else:
        zone_letter = force_zone_letter

    lon_rad = mathlib.radians(longitude)
    central_lon = zone_number_to_central_longitude(zone_number)
    central_lon_rad = mathlib.radians(central_lon)

    n = R / mathlib.sqrt(1 - E * lat_sin ** 2)
    c = E_P2 * lat_cos ** 2

    a = lat_cos * (lon_rad - central_lon_rad)
    a2 = a * a
    a3 = a2 * a
    a4 = a3 * a
    a5 = a4 * a
    a6 = a5 * a

    m = R * (
        M1 * lat_rad -
        M2 * mathlib.sin(2 * lat_rad) +
        M3 * mathlib.sin(4 * lat_rad) -
        M4 * mathlib.sin(6 * lat_rad)
    )

    easting = (
        K0 *
        n *
        (
            a +
            a3 / 6 * (1 - lat_tan2 + c) +
            a5 / 120 * (5 - 18 * lat_tan2 + lat_tan4 + 72 * c - 58 * E_P2)
        ) + 500000
    )

    northing = K0 * (
        m +
        n *
        lat_tan * (
            a2 / 2 + a4 / 24 * (5 - lat_tan2 + 9 * c + 4 * c ** 2) +
            a6 / 720 * (61 - 58 * lat_tan2 + lat_tan4 + 600 * c - 330 * E_P2)
        )
    )

    if mixed_signs(latitude):
        raise ValueError("latitudes must all have the same sign")
    elif negative(latitude):
        northing += 10000000

    return easting, northing, zone_number, zone_letter


def latitude_to_zone_letter(latitude):
    # If the input is a numpy array, just use the first element
    # User responsibility to make sure that all points are in one zone
    if use_numpy and isinstance(latitude, mathlib.ndarray):
        latitude = latitude.flat[0]

    if -80 <= latitude <= 84:
        return ZONE_LETTERS[int(latitude + 80) >> 3]
    else:
        return None


def latlon_to_zone_number(latitude, longitude):
    # If the input is a numpy array, just use the first element
    # User responsibility to make sure that all points are in one zone
    if use_numpy:
        if isinstance(latitude, mathlib.ndarray):
            latitude = latitude.flat[0]
        if isinstance(longitude, mathlib.ndarray):
            longitude = longitude.flat[0]

    if 56 <= latitude < 64 and 3 <= longitude < 12:
        return 32

    if 72 <= latitude <= 84 and longitude >= 0:
        if longitude < 9:
            return 31
        elif longitude < 21:
            return 33
        elif longitude < 33:
            return 35
        elif longitude < 42:
            return 37

    return int((longitude + 180) / 6) + 1


def zone_number_to_central_longitude(zone_number):
    return (zone_number - 1) * 6 - 180 + 3



import collections
import functools


@functools.lru_cache(maxsize=2)
def neighbor_flips(partition):
    """The neighbors of the flips in a partition"""
    return {
        tuple(sorted((node, neighbor)))
        for node in partition.flips
        for neighbor in partition.graph.neighbors(node)
    }


def create_flow():
    return {'in': set(), 'out': set()}


@functools.lru_cache(maxsize=2)
def flows_from_changes(old_partition, new_partition):
    flows = collections.defaultdict(create_flow)
    for node, target in new_partition.flips.items():
        source = old_partition.assignment.mapping[node]
        if source != target:
            flows[target]['in'].add(node)
            flows[source]['out'].add(node)
    return flows


def on_flow(initializer, alias):
    """
    Use this decorator to create an updater that responds to flows of nodes
    between parts of the partition.

    Decorate a function that takes:
    - The partition
    - The previous value of the updater on a fixed part P_i
    - The new nodes that are just joining P_i at this step
    - The old nodes that are just leaving P_i at this step
    and returns:
    - The new value of the updater for the fixed part P_i.

    This will create an updater whose values are dictionaries of the
    form `{part: <value of the given function on the part>}`.

    The initializer, by contrast, should take the entire partition and
    return the entire `{part: <value>}` dictionary.

    Example:

    .. code-block:: python

        @on_flow(initializer, alias='my_updater')
        def my_updater(partition, previous, new_nodes, old_nodes):
            # return new value for the part
    """
    def decorator(function):
        @functools.wraps(function)
        def wrapped(partition, previous=None):
            if partition.parent is None:
                return initializer(partition)

            if previous is None:
                previous = partition.parent[alias]

            new_values = previous.copy()

            for part, flow in partition.flows.items():
                new_values[part] = function(partition, previous[part], flow['in'], flow['out'])

            return new_values
        return wrapped
    return decorator


def compute_edge_flows(partition):
    edge_flows = collections.defaultdict(create_flow)
    assignment = partition.assignment
    old_assignment = partition.parent.assignment

    for (node, neighbor) in neighbor_flips(partition):
        edge = (node, neighbor)

        old_source = old_assignment.mapping[node]
        old_target = old_assignment.mapping[neighbor]

        new_source = assignment.mapping[node]
        new_target = assignment.mapping[neighbor]

        cut = new_source != new_target
        was_cut = old_source != old_target

        if not cut and was_cut:
            edge_flows[old_target]['out'].add(edge)
            edge_flows[old_source]['out'].add(edge)
        elif cut and not was_cut:
            edge_flows[new_target]['in'].add(edge)
            edge_flows[new_source]['in'].add(edge)
        elif cut and was_cut:
            # If an edge was cut and still is cut, we need to make sure the
            # edge is listed under the correct parts.
            no_longer_incident_parts = {old_target, old_source} - \
                                        {new_target, new_source}
            for part in no_longer_incident_parts:
                edge_flows[part]['out'].add(edge)

            newly_incident_parts = {new_target, new_source} - {old_target, old_source}
            for part in newly_incident_parts:
                edge_flows[part]['in'].add(edge)
    return edge_flows


def on_edge_flow(initializer, alias):
    """
    Use this decorator to create an updater that responds to flows of cut
    edges between parts of the partition.

    Decorate a function that takes:
    - The partition
    - The previous value of the updater for a fixed part P_i
    - The new cut edges that are just joining P_i at this step
    - The old cut edges that are just leaving P_i at this step
    and returns:
    - The new value of the updater for the fixed part P_i.

    This will create an updater whose values are dictionaries of the
    form `{part: <value of the given function on the part>}`.

    The initializer, by contrast, should take the entire partition and
    return the entire `{part: <value>}` dictionary.

    Example:

    .. code-block:: python

        @on_edge_flow(initializer, alias='my_updater')
        def my_updater(partition, previous, new_edges, old_edges):
            # return new value of the part
    """
    def decorator(f):
        @functools.wraps(f)
        def wrapper(partition):
            if not partition.parent:
                return initializer(partition)
            edge_flows = partition.edge_flows
            previous = partition.parent[alias]

            new_values = previous.copy()
            for part in partition.edge_flows:
                new_values[part] = f(partition, previous[part],
                                     new_edges=edge_flows[part]['in'],
                                     old_edges=edge_flows[part]['out'])
            return new_values
        return wrapper
    return decorator



import networkx as nx
from numpy import linalg as LA
from ..random import random


def spectral_cut(graph, part_labels, weight_type, lap_type):
    """Spectral cut function.

    Uses the signs of the elements in the Fiedler vector of a graph to
    partition into two components.

    """

    nlist = list(graph.nodes())
    n = len(nlist)

    if weight_type == "random":
        for edge in graph.edge_indices:
            graph.edges[edge]["weight"] = random.random()

    if lap_type == "normalized":
        LAP = (nx.normalized_laplacian_matrix(graph)).todense()

    else:
        LAP = (nx.laplacian_matrix(graph)).todense()

    NLMva, NLMve = LA.eigh(LAP)
    NFv = NLMve[:, 1]
    xNFv = [NFv.item(x) for x in range(n)]

    node_color = [xNFv[x] > 0 for x in range(n)]

    clusters = {nlist[x]: part_labels[node_color[x]] for x in range(n)}

    return clusters


def spectral_recom(partition, weight_type=None, lap_type="normalized"):
    """Spectral ReCom proposal.

    Uses spectral clustering to bipartition a subgraph of the original graph
    formed by merging the nodes corresponding to two adjacent districts.

    Example usage::

        from functools import partial
        from gerrychain import MarkovChain
        from gerrychain.proposals import recom

        # ...define constraints, accept, partition, total_steps here...


        proposal = partial(
            spectral_recom, weight_type=None, lap_type="normalized"
        )

        chain = MarkovChain(proposal, constraints, accept, partition, total_steps)

    """

    edge = random.choice(tuple(partition["cut_edges"]))
    parts_to_merge = (partition.assignment.mapping[edge[0]], partition.assignment.mapping[edge[1]])

    subgraph = partition.graph.subgraph(
        partition.parts[parts_to_merge[0]] | partition.parts[parts_to_merge[1]]
    )

    flips = spectral_cut(
        subgraph,
        parts_to_merge,
        weight_type,
        lap_type
    )

    return partition.flip(flips)



from ..random import random


def propose_any_node_flip(partition):
    """Flip a random node (not necessarily on the boundary) to a random part
    """

    node = random.choice(tuple(partition.graph))
    newpart = random.choice(tuple(partition.parts))

    return partition.flip({node: newpart})


def propose_flip_every_district(partition):
    """Proposes a random boundary flip for each district in the partition.

    :param partition: The current partition to propose a flip from.
    :return: a proposed next `~gerrychain.Partition`
    """
    flips = dict()

    for dist_edges in partition["cut_edges_by_part"].values():
        edge = random.choice(tuple(dist_edges))

        index = random.choice((0, 1))
        flipped_node, other_node = edge[index], edge[1 - index]
        flip = {flipped_node: partition.assignment.mapping[other_node]}

        flips.update(flip)

    return partition.flip(flips)


def propose_chunk_flip(partition):
    """Chooses a random boundary node and proposes to flip it and all of its neighbors

    :param partition: The current partition to propose a flip from.
    :return: a proposed next `~gerrychain.Partition`
    """
    flips = dict()

    edge = random.choice(tuple(partition["cut_edges"]))
    index = random.choice((0, 1))

    flipped_node = edge[index]

    valid_flips = [
        nbr
        for nbr in partition.graph.neighbors(flipped_node)
        if partition.assignment.mapping[nbr] != partition.assignment.mapping[flipped_node]
    ]

    for flipped_neighbor in valid_flips:
        flips.update({flipped_neighbor: partition.assignment.mapping[flipped_node]})

    return partition.flip(flips)


def propose_random_flip(partition):
    """Proposes a random boundary flip from the partition.

    :param partition: The current partition to propose a flip from.
    :return: a proposed next `~gerrychain.Partition`
    """
    if len(partition["cut_edges"]) == 0:
        return partition
    edge = random.choice(tuple(partition["cut_edges"]))
    index = random.choice((0, 1))
    flipped_node, other_node = edge[index], edge[1 - index]
    flip = {flipped_node: partition.assignment.mapping[other_node]}
    return partition.flip(flip)


def slow_reversible_propose_bi(partition):
    """Proposes a random boundary flip from the partition in a reversible fasion
    for bipartitions by selecting a boundary node at random and uniformly picking
    one of its neighboring parts. For k-partitions this is not uniform since there
    might be multiple parts next to a single node.

    Temporary version until we make an updater for this set.

    :param partition: The current partition to propose a flip from.
    :return: a proposed next `~gerrychain.Partition`
    """

    b_nodes = {x[0] for x in partition["cut_edges"]}.union({x[1] for x in partition["cut_edges"]})

    flip = random.choice(list(b_nodes))
    neighbor_assignments = list(set([partition.assignment.mapping[neighbor] for neighbor
                                in partition.graph.neighbors(flip)]))
    neighbor_assignments.remove(partition.assignment.mapping[flip])
    flips = {flip: random.choice(neighbor_assignments)}

    return partition.flip(flips)


flip = propose_random_flip


def slow_reversible_propose(partition):
    """Proposes a random boundary flip from the partition in a reversible fasion
    by selecting uniformly from the (node, flip) pairs.

    Temporary version until we make an updater for this set.

    :param partition: The current partition to propose a flip from.
    :return: a proposed next `~gerrychain.Partition`
    """

    b_nodes = {(x[0], partition.assignment.mapping[x[1]]) for x in partition["cut_edges"]
               }.union({(x[1], partition.assignment.mapping[x[0]]) for x in partition["cut_edges"]})

    flip = random.choice(list(b_nodes))
    return partition.flip({flip[0]: flip[1]})



from heapq import heappop, heappush
from itertools import count

import networkx as nx

from ..random import random
from .bounds import SelfConfiguringLowerBound


def are_reachable(G, source, avoid, targets):
    """Check that source can reach targets while avoiding given edges.
    This function is a modified form of NetworkX's ``_dijkstra_multisource()``.

    :param G: NetworkX graph.
    :param source: Starting node.
    :param weight: Function with (u, v, data) input that returns that edges weight.
    :param targets: Nodes required to find.
    :return:
        A mapping from node to shortest distance to that node from one
        of the source nodes.
    :rtype: dict
    """
    G_succ = G._succ if G.is_directed() else G._adj

    push = heappush
    pop = heappop
    dist = {}  # dictionary of final distances
    seen = {}
    # fringe is heapq with 3-tuples (distance,c,node)
    # use the count c to avoid comparing nodes (may not be able to)
    c = count()
    fringe = []

    seen[source] = 0
    push(fringe, (0, next(c), source))

    while not all(t in seen for t in targets) and fringe:
        (d, _, v) = pop(fringe)
        if v in dist:
            continue  # already searched this node.
        dist[v] = d
        for u, e in G_succ[v].items():
            if avoid(v, u, e):
                continue

            vu_dist = dist[v] + 1
            if u not in seen or vu_dist < seen[u]:
                seen[u] = vu_dist
                push(fringe, (vu_dist, next(c), u))

    return all(t in seen for t in targets)


def single_flip_contiguous(partition):
    """Check if swapping the given node from its old assignment disconnects the
    old assignment class.

    :param partition: The proposed next :class:`~gerrychain.partition.Partition`

    :return: whether the partition is contiguous
    :rtype: bool

    We assume that `removed_node` belonged to an assignment class that formed a
    connected subgraph. To see if its removal left the subgraph connected, we
    check that the neighbors of the removed node are still connected through
    the changed graph.

    """
    parent = partition.parent
    flips = partition.flips
    if not flips or not parent:
        return contiguous(partition)

    graph = partition.graph
    assignment = partition.assignment

    def partition_edge_avoid(start_node, end_node, edge_attrs):
        """Compute the district edge weight, which is 1 if the nodes have the same
        assignment, and infinity otherwise.
        """
        if assignment.mapping[start_node] != assignment.mapping[end_node]:
            # Fun fact: networkx actually refuses to take edges with None
            # weight.
            return True

        return False

    for changed_node in flips:
        old_assignment = partition.parent.assignment.mapping[changed_node]

        old_neighbors = [
            node
            for node in graph.neighbors(changed_node)
            if assignment.mapping[node] == old_assignment
        ]

        if not old_neighbors:
            # Under our assumptions, if there are no old neighbors, then the
            # old_assignment district has vanished. It is trivially connected.

            # However, we actually don't want any districts to disappear because
            # it ends up breaking a lot of our other updaters. So we consider
            # the empty district to be disconnected.
            return False

        start_neighbor = random.choice(old_neighbors)

        connected = are_reachable(
            graph, start_neighbor, partition_edge_avoid, old_neighbors
        )

        if not connected:
            return False

    # All neighbors of all changed nodes are connected, so the new graph is
    # connected.
    return True


def affected_parts(partition):
    """Returns the set of IDs of all parts that gained or lost a node during the last
    flip.
    """
    flips = partition.flips
    parent = partition.parent

    if flips is None:
        return partition.parts

    affected = set()
    for node, part in flips.items():
        affected.add(part)
        affected.add(parent.assignment.mapping[node])

    return affected


def contiguous(partition):
    """Check if the parts of a partition are connected using :func:`networkx.is_connected`.

    :param partition: The proposed next :class:`~gerrychain.partition.Partition`

    :return: whether the partition is contiguous
    :rtype: bool
    """
    return all(
        nx.is_connected(partition.subgraphs[part]) for part in affected_parts(partition)
    )


def contiguous_bfs(partition):
    """Checks that a given partition's parts are connected as graphs using a simple
    breadth-first search.

    :param partition: Instance of Partition
    :return: Whether the parts of this partition are connected
    :rtype: bool
    """
    parts_to_check = affected_parts(partition)

    # Generates a subgraph for each district and perform a BFS on it
    # to check connectedness.
    for part in parts_to_check:
        adj = nx.to_dict_of_lists(partition.subgraphs[part])
        if _bfs(adj) is False:
            return False

    return True


def number_of_contiguous_parts(partition):
    """Return the number of non-connected assignment subgraphs.

    :param partition: Instance of Partition; contains connected components.
    :return: number of contiguous districts
    :rtype: int
    """
    parts = partition.assignment.parts
    return sum(1 for part in parts if nx.is_connected(partition.subgraphs[part]))


def contiguous_components(partition):
    """Return then connected components of each of the subgraphs of the parts
    of the partition.

    :param partition: Instance of Partition; contains connected components.
    :return: dictionary mapping each part ID to a list holding the connected
        subgraphs of that part of the partition
    :rtype: dict
    """
    return {
        part: [subgraph.subgraph(nodes) for nodes in nx.connected_components(subgraph)]
        for part, subgraph in partition.subgraphs.items()
    }


no_more_discontiguous = SelfConfiguringLowerBound(number_of_contiguous_parts)


def _bfs(graph):
    """Performs a breadth-first search on the provided graph and returns true or
    false depending on whether the graph is connected.

    :param graph: Dict-of-lists; an adjacency matrix.
    :return: is this graph connected?
    :rtype: bool
    """
    q = [next(iter(graph))]
    visited = set()
    total_vertices = len(graph)

    # Check if the district has a single vertex. If it does, then simply return
    # `True`, as it's trivially connected.
    if total_vertices <= 1:
        return True

    # bfs!
    while len(q) > 0:
        current = q.pop(0)
        neighbors = graph[current]

        for neighbor in neighbors:
            if neighbor not in visited:
                visited.add(neighbor)
                q += [neighbor]

    return total_vertices == len(visited)



"""Updaters that compute spanning tree statistics."""
import math
import numpy
import networkx


def _num_spanning_trees_in_district(partition, district):
    """Given a district ID, returns the number of spanning trees in the
    subgraph of self corresponding to the district.

    Uses Kirchoff's theorem to compute the number of spanning trees.

    :param partition: :class:`gerrychain.Partition`
    :param district: A district label (part) in the partition.
    :return: The number of spanning trees in the subgraph of the
    partition corresponding to district
    """
    graph = partition.subgraphs[district]
    laplacian = networkx.laplacian_matrix(graph)
    L = numpy.delete(numpy.delete(laplacian.todense(), 0, 0), 1, 1)
    return math.exp(numpy.linalg.slogdet(L)[1])


def num_spanning_trees(partition):
    """Returns the number of spanning trees in each part (district) of a partition."""
    return {
        part: _num_spanning_trees_in_district(partition, part)
        for part in partition.parts
    }



from .proposals import *
from .tree_proposals import recom, reversible_recom, ReCom
from .spectral_proposals import spectral_recom

__all__ = ["recom", "reversible_recom", "spectral_recom", "propose_chunk_flip", "propose_random_flip"]



from ..updaters import CountySplit
from .bounds import Bounds
import numpy


class Validator:
    """A single callable for checking that a partition passes a collection of
    constraints. Intended to be passed as the ``is_valid`` parameter when
    instantiating :class:`~gerrychain.MarkovChain`.

    This class is meant to be called as a function after instantiation; its
    return is ``True`` if all validators pass, and ``False`` if any one fails.

    Example usage::

        is_valid = Validator([constraint1, constraint2, constraint3])
        chain = MarkovChain(proposal, is_valid, accept, initial_state, total_steps)
    """

    def __init__(self, constraints):
        """
        :param constraints: List of validator functions that will check partitions.
        """
        self.constraints = constraints

    def __call__(self, partition):
        """Determine if the given partition is valid.

        :param partition: :class:`Partition` class to check.

        """
        # check each constraint function and fail when a constraint test fails
        for constraint in self.constraints:
            is_valid = constraint(partition)
            # Coerce NumPy booleans
            if isinstance(is_valid, numpy.bool_):
                is_valid = bool(is_valid)

            if is_valid is False:
                return False
            elif is_valid is True:
                pass
            else:
                raise TypeError(
                    "Constraint {} returned a non-boolean.".format(repr(constraint))
                )

        # all constraints are satisfied
        return True


def within_percent_of_ideal_population(
    initial_partition, percent=0.01, pop_key="population"
):
    """Require that all districts are within a certain percent of "ideal" (i.e.,
    uniform) population.

    Ideal population is defined as "total population / number of districts."

    :param initial_partition: Starting partition from which to compute district information.
    :param percent: (optional) Allowed percentage deviation. Default is 1%.
    :param pop_key: (optional) The name of the population
        :class:`Tally <gerrychain.updaters.Tally>`. Default is ``"population"``.
    :return: A :class:`.Bounds` constraint on the population attribute identified
        by ``pop_key``.
    """

    def population(partition):
        return partition[pop_key].values()

    number_of_districts = len(initial_partition[pop_key].keys())
    total_population = sum(initial_partition[pop_key].values())
    ideal_population = total_population / number_of_districts
    bounds = ((1 - percent) * ideal_population, (1 + percent) * ideal_population)

    return Bounds(population, bounds=bounds)


def deviation_from_ideal(partition, attribute="population"):
    """Computes the deviation of the given ``attribute`` from exact equality
    among parts of the partition. Usually ``attribute`` is the population, and
    this function is used to compute how far a districting plan is from exact population
    equality.

    By "deviation" we mean ``(actual_value - ideal)/ideal`` (not the absolute value).

    :param partition: A partition.
    :param attribute: (optional) The :class:`Tally <gerrychain.updaters.Tally>` to
        compute deviation for. Default is ``"population"``.
    :return: dictionary from parts to their deviation
    """
    number_of_districts = len(partition[attribute].keys())
    total = sum(partition[attribute].values())
    ideal = total / number_of_districts

    return {
        part: (value - ideal) / ideal for part, value in partition[attribute].items()
    }


def districts_within_tolerance(partition, attribute_name="population", percentage=0.1):
    """Check if all districts are within a certain percentage of the "smallest"
    district, as defined by the given attribute.

    :param partition: partition class instance
    :param attrName: string that is the name of an updater in partition
    :param percentage: what percent difference is allowed
    :return: whether the districts are within specified tolerance
    :rtype: bool
    """
    if percentage >= 1:
        percentage *= 0.01

    values = partition[attribute_name].values()
    max_difference = max(values) - min(values)

    within_tolerance = max_difference <= percentage * min(values)
    return within_tolerance


def refuse_new_splits(partition_county_field):
    """Refuse all proposals that split a county that was previous unsplit.

    :param partition_county_field: Name of field for county information generated by
        :func:`.county_splits`.
    """

    def _refuse_new_splits(partition):
        for county_info in partition[partition_county_field].values():
            if county_info.split == CountySplit.NEW_SPLIT:
                return False

        return True

    return _refuse_new_splits


def no_vanishing_districts(partition):
    """Require that no districts be completely consumed."""
    if not partition.parent:
        return True
    return all(len(part) > 0 for part in partition.assignment.parts.values())



import collections

from .flows import on_flow
from .cut_edges import on_edge_flow


def boundary_nodes(partition, alias="boundary_nodes"):
    if partition.parent:
        return partition.parent[alias]
    return {
        node
        for node in partition.graph.nodes
        if partition.graph.nodes[node]["boundary_node"]
    }


def initialize_exterior_boundaries_as_a_set(partition):
    part_boundaries = collections.defaultdict(set)
    for node in partition["boundary_nodes"]:
        part_boundaries[partition.assignment.mapping[node]].add(node)
    return part_boundaries


@on_flow(initialize_exterior_boundaries_as_a_set, alias="exterior_boundaries_as_a_set")
def exterior_boundaries_as_a_set(partition, previous, inflow, outflow):
    graph_boundary = partition["boundary_nodes"]
    return (previous | (inflow & graph_boundary)) - outflow


def initialize_exterior_boundaries(partition):
    graph_boundary = partition["boundary_nodes"]
    boundaries = collections.defaultdict(lambda: 0)
    for node in graph_boundary:
        part = partition.assignment.mapping[node]
        boundaries[part] += partition.graph.nodes[node]["boundary_perim"]
    return boundaries


@on_flow(initialize_exterior_boundaries, alias="exterior_boundaries")
def exterior_boundaries(partition, previous, inflow, outflow):
    graph_boundary = partition["boundary_nodes"]
    added_perimeter = sum(
        partition.graph.nodes[node]["boundary_perim"]
        for node in inflow & graph_boundary
    )
    removed_perimeter = sum(
        partition.graph.nodes[node]["boundary_perim"]
        for node in outflow & graph_boundary
    )
    return previous + added_perimeter - removed_perimeter


def initialize_interior_boundaries(partition):
    return {
        part: sum(
            partition.graph.edges[edge]["shared_perim"]
            for edge in partition["cut_edges_by_part"][part]
        )
        for part in partition.parts
    }


@on_edge_flow(initialize_interior_boundaries, alias="interior_boundaries")
def interior_boundaries(partition, previous, new_edges, old_edges):
    added_perimeter = sum(
        partition.graph.edges[edge]["shared_perim"] for edge in new_edges
    )
    removed_perimeter = sum(
        partition.graph.edges[edge]["shared_perim"] for edge in old_edges
    )
    return previous + added_perimeter - removed_perimeter


def flips(partition):
    return partition.flips


def perimeter_of_part(partition, part):
    """
    Totals up the perimeter of the part in the partition.
    Requires that 'boundary_perim' be a node attribute, 'shared_perim' be an edge
    attribute, 'cut_edges' be an updater, and 'exterior_boundaries' be an updater.
    """
    exterior_perimeter = partition["exterior_boundaries"][part]
    interior_perimeter = partition["interior_boundaries"][part]

    return exterior_perimeter + interior_perimeter


def perimeter(partition):
    return {part: perimeter_of_part(partition, part) for part in partition.parts}



import math

from gerrychain.updaters.tally import DataTally
import gerrychain.metrics.partisan as pm


class Election:
    """Represents the data of one election, with races conducted in each part of
    the partition.

    As we vary the districting plan, we can use the same node-level vote totals
    to tabulate hypothetical elections. To do this manually with tallies, we would
    have to maintain tallies for each party, as well as the total number of votes,
    and then compute the electoral results and percentages from scratch every time.
    To make this simpler, this class provides an :class:`ElectionUpdater` to manage
    these tallies. The updater returns an :class:`ElectionResults` class giving
    a convenient view of the election results, with methods like
    :meth:`~ElectionResults.wins` or :meth:`~ElectionResults.percent` for common queries
    the user might make on election results.

    Example usage::

        # Assuming your nodes have attributes "2008_D", "2008_R"
        # with (for example) 2008 senate election vote totals
        election = Election(
            "2008 Senate",
            {"Democratic": "2008_D", "Republican": "2008_R"},
            alias="2008_Sen"
        )

        # Assuming you already have a graph and assignment:
        partition = Partition(
            graph,
            assignment,
            updaters={"2008_Sen": election}
        )

        # The updater returns an ElectionResults instance, which
        # we can use (for example) to see how many seats a given
        # party would win in this partition using this election's
        # vote distribution:
        partition["2008_Sen"].wins("Republican")

    """

    def __init__(self, name, parties_to_columns, alias=None):
        """
        :param name: The name of the election. (e.g. "2008 Presidential")
        :param parties_to_columns: A dictionary matching party names to their
            data columns, either as actual columns (list-like, indexed by nodes)
            or as string keys for the node attributes that hold the party's
            vote totals. Or, a list of strings which will serve as both
            the party names and the node attribute keys.
        :param alias: (optional) Alias that the election is registered under
            in the Partition's dictionary of updaters.
        """
        self.name = name

        if alias is None:
            alias = name
        self.alias = alias

        if isinstance(parties_to_columns, dict):
            self.parties = list(parties_to_columns.keys())
            self.columns = list(parties_to_columns.values())
            self.parties_to_columns = parties_to_columns
        elif isinstance(parties_to_columns, list):
            self.parties = parties_to_columns
            self.columns = parties_to_columns
            self.parties_to_columns = dict(zip(self.parties, self.columns))
        else:
            raise TypeError("Election expects parties_to_columns to be a dict or list")

        self.tallies = {
            party: DataTally(self.parties_to_columns[party], party)
            for party in self.parties
        }

        self.updater = ElectionUpdater(self)

    def __str__(self):
        return "Election '{}' with vote totals for parties {} from columns {}.".format(
            self.name, str(self.parties), str(self.columns)
        )

    def __repr__(self):
        return "Election(parties={}, columns={}, alias={})".format(
            str(self.parties), str(self.columns), str(self.alias)
        )

    def __call__(self, *args, **kwargs):
        return self.updater(*args, **kwargs)


class ElectionUpdater:
    """
    The updater for computing the election results in each part of the partition after
    each step in the Markov chain. The actual results are returned to the user as
    an :class:`ElectionResults` instance.
    """

    def __init__(self, election):
        self.election = election

    def __call__(self, partition):
        previous_totals_for_party = self.get_previous_values(partition)
        parties = self.election.parties
        tallies = self.election.tallies

        counts = {
            party: tallies[party](partition, previous=previous_totals_for_party[party])
            for party in parties
        }

        return ElectionResults(self.election, counts, races=partition.parts)

    def get_previous_values(self, partition):
        parent = partition.parent
        if parent is None:
            previous_totals_for_party = {party: None for party in self.election.parties}
        else:
            previous_totals_for_party = partition.parent[
                self.election.alias
            ].totals_for_party
        return previous_totals_for_party


def get_percents(counts, totals):
    return {
        part: counts[part] / totals[part] if totals[part] > 0 else math.nan
        for part in totals
    }


class ElectionResults:
    """
    Represents the results of an election. Provides helpful methods to answer
    common questions you might have about an election (Who won? How many seats?, etc.).
    """

    def __init__(self, election, counts, races):
        self.election = election
        self.totals_for_party = counts
        self.races = races

        self.totals = {
            race: sum(counts[party][race] for party in self.election.parties)
            for race in self.races
        }

        self.percents_for_party = {
            party: get_percents(counts[party], self.totals)
            for party in election.parties
        }

    def __str__(self):
        results_by_part = "\n".join(
            format_part_results(self.percents_for_party, part) for part in self.totals
        )
        return "Election Results for {name}\n{results}".format(
            name=self.election.name, results=results_by_part
        )

    def seats(self, party):
        """
        Returns the number of races that ``party`` won.
        """
        return sum(self.won(party, race) for race in self.races)

    def wins(self, party):
        """
        An alias for :meth:`seats`.
        """
        return self.seats(party)

    def percent(self, party, race=None):
        """
        Returns the percentage of the vote that ``party`` received in a given race
        (part of the partition). If ``race`` is omitted, returns the overall vote
        share of ``party``.

        :param party: Party ID.
        :param race: ID of the part of the partition whose votes we want to tally.
        """
        if race is not None:
            return self.percents_for_party[party][race]
        return sum(self.votes(party)) / sum(self.totals[race] for race in self.races)

    def percents(self, party):
        """
        :param party: The party
        :return: The tuple of the percentage of votes that ``party`` received
            in each part of the partition
        """
        return tuple(self.percents_for_party[party][race] for race in self.races)

    def count(self, party, race=None):
        """
        Returns the total number of votes that ``party`` received in a given race
        (part of the partition). If ``race`` is omitted, returns the overall vote
        total of ``party``.

        :param party: Party ID.
        :param race: ID of the part of the partition whose votes we want to tally.
        """
        if race is not None:
            return self.totals_for_party[party][race]
        return sum(self.totals_for_party[party][race] for race in self.races)

    def counts(self, party):
        """
        :param party: Party ID
        :return: tuple of the total votes cast for ``party`` in each part of
            the partition
        """
        return tuple(self.totals_for_party[party][race] for race in self.races)

    def votes(self, party):
        """
        An alias for :meth:`counts`.
        """
        return self.counts(party)

    def won(self, party, race):
        """
        Answers "Did ``party`` win the race in part ``race``?" with ``True`` or ``False``.
        """
        return all(
            self.totals_for_party[party][race] >= self.totals_for_party[opponent][race]
            for opponent in self.election.parties
        )

    def total_votes(self):
        return sum(self.totals.values())

    def mean_median(self):
        """
        Computes the mean-median score for this ElectionResults object.
        """
        return pm.mean_median(self)

    def mean_thirdian(self):
        """
        Computes the mean-thirdian score for this ElectionResults object.
        """
        return pm.mean_thirdian(self)

    def efficiency_gap(self):
        """
        Computes the efficiency gap for this ElectionResults object.
        """
        return pm.efficiency_gap(self)

    def partisan_bias(self):
        """
        Computes the partisan bias for this ElectionResults object.
        """
        return pm.partisan_bias(self)

    def partisan_gini(self):
        """
        Computes the Gini score for this ElectionResults object.
        """
        return pm.partisan_gini(self)


def format_part_results(percents_for_party, part):
    heading = "{part}:\n".format(part=str(part))
    body = "\n".join(
        "  {party}: {percent}".format(
            party=str(party), percent=round(percents_for_party[party][part], 4)
        )
        for party in percents_for_party
    )
    return heading + body



from .random import random


def always_accept(partition):
    return True


def cut_edge_accept(partition):
    """Always accepts the flip if the number of cut_edges increases.
    Otherwise, uses the Metropolis criterion to decide.

    :param partition: The current partition to accept a flip from.
    :return: True if accepted, False to remain in place

    """
    bound = 1

    if partition.parent is not None:
        bound = min(1, len(partition.parent["cut_edges"]) / len(partition["cut_edges"]))

    return random.random() < bound



import math

import networkx

from gerrychain.partition import Partition
from gerrychain.graph import Graph
from gerrychain.updaters import (
    Tally,
    boundary_nodes,
    cut_edges,
    cut_edges_by_part,
    exterior_boundaries,
    interior_boundaries,
    perimeter,
)
from gerrychain.metrics import polsby_popper


class Grid(Partition):
    """
    The :class:`Grid` class represents a grid partitioned into districts.
    It is useful for running little experiments with GerryChain without needing to do
    any data processing or cleaning to get started.

    Example usage::

        grid = Grid((10,10))

    The nodes of ``grid.graph`` are labelled by tuples ``(i,j)``, for ``0 <= i <= 10``
    and ``0 <= j <= 10``. Each node has an ``area`` of 1 and each edge has ``shared_perim`` 1.
    """

    default_updaters = {
        "cut_edges": cut_edges,
        "population": Tally("population"),
        "perimeter": perimeter,
        "exterior_boundaries": exterior_boundaries,
        "interior_boundaries": interior_boundaries,
        "boundary_nodes": boundary_nodes,
        "area": Tally("area", alias="area"),
        "polsby_popper": polsby_popper,
        "cut_edges_by_part": cut_edges_by_part,
    }

    def __init__(
        self,
        dimensions=None,
        with_diagonals=False,
        assignment=None,
        updaters=None,
        parent=None,
        flips=None,
    ):
        """
        :param dimensions: tuple (m,n) of the desired dimensions of the grid.
        :param with_diagonals: (optional, defaults to False) whether to include diagonals
            as edges of the graph (i.e., whether to use 'queen' adjacency rather than
            'rook' adjacency).
        :param assignment: (optional) dict matching nodes to their districts. If not
            provided, partitions the grid into 4 quarters of roughly equal size.
        :param updaters: (optional) dict matching names of attributes of the Partition
            to functions that compute their values. If not provided, the Grid
            configures the cut_edges updater for convenience.
        """
        if dimensions:
            self.dimensions = dimensions
            graph = Graph.from_networkx(create_grid_graph(dimensions, with_diagonals))

            if not assignment:
                thresholds = tuple(math.floor(n / 2) for n in self.dimensions)
                assignment = {
                    node: color_quadrants(node, thresholds) for node in graph.nodes
                }

            if not updaters:
                updaters = dict()
            updaters.update(self.default_updaters)

            super().__init__(graph, assignment, updaters)
        elif parent:
            self.dimensions = parent.dimensions
            super().__init__(parent=parent, flips=flips)
        else:
            raise Exception("Not a good way to create a Partition")

    def __str__(self):
        rows = self.as_list_of_lists()
        return "\n".join(["".join([str(x) for x in row]) for row in rows]) + "\n"

    def __repr__(self):
        dims = "x".join(str(d) for d in self.dimensions)
        number_of_parts = len(self.parts)
        s = "s" if number_of_parts > 1 else ""
        return "{} Grid\nPartitioned into {} part{}".format(dims, number_of_parts, s)

    def as_list_of_lists(self):
        """
        Returns the grid as a list of lists (like a matrix), where the (i,j)th
        entry is the assigned district of the node in position (i,j) on the
        grid.
        """
        m, n = self.dimensions
        return [[self.assignment.mapping[(i, j)] for i in range(m)] for j in range(n)]


def create_grid_graph(dimensions, with_diagonals):
    if len(dimensions) != 2:
        raise ValueError("Expected two dimensions.")
    m, n = dimensions
    graph = networkx.generators.lattice.grid_2d_graph(m, n)

    networkx.set_edge_attributes(graph, 1, "shared_perim")

    if with_diagonals:
        nw_to_se = [
            ((i, j), (i + 1, j + 1)) for i in range(m - 1) for j in range(n - 1)
        ]
        sw_to_ne = [
            ((i, j + 1), (i + 1, j)) for i in range(m - 1) for j in range(n - 1)
        ]
        diagonal_edges = nw_to_se + sw_to_ne
        graph.add_edges_from(diagonal_edges)
        for edge in diagonal_edges:
            graph.edges[edge]["shared_perim"] = 0

    networkx.set_node_attributes(graph, 1, "population")
    networkx.set_node_attributes(graph, 1, "area")

    tag_boundary_nodes(graph, dimensions)

    return graph


def give_constant_attribute(graph, attribute, value):
    for node in graph.nodes:
        graph.nodes[node][attribute] = value


def tag_boundary_nodes(graph, dimensions):
    m, n = dimensions
    for node in graph.nodes:
        if node[0] in [0, m - 1] or node[1] in [0, n - 1]:
            graph.nodes[node]["boundary_node"] = True
            graph.nodes[node]["boundary_perim"] = get_boundary_perim(node, dimensions)
        else:
            graph.nodes[node]["boundary_node"] = False


def get_boundary_perim(node, dimensions):
    m, n = dimensions
    if node in [(0, 0), (m - 1, 0), (0, n - 1), (m - 1, n - 1)]:
        return 2
    elif node[0] in [0, m - 1] or node[1] in [0, n - 1]:
        return 1
    else:
        return 0


def color_half(node, threshold=5):
    x = node[0]
    return 0 if x <= threshold else 1


def color_quadrants(node, thresholds):
    x, y = node
    x_color = 0 if x < thresholds[0] else 1
    y_color = 0 if y < thresholds[1] else 2
    return x_color + y_color


def grid_size(parition):
    """ This is a hardcoded population function
    for the grid class"""

    L = parition.as_list_of_lists()
    permit = [3, 4, 5]

    sizes = [0, 0, 0, 0]

    for i in range(len(L)):
        for j in range(len(L[0])):
            sizes[L[i][j]] += 1

    return all(x in permit for x in sizes)



import collections
from enum import Enum

CountyInfo = collections.namedtuple("CountyInfo", "split nodes contains")


class CountySplit(Enum):
    NOT_SPLIT = 0
    NEW_SPLIT = 1
    OLD_SPLIT = 2


def county_splits(partition_name, county_field_name):
    """Track county splits.

    :param partition_name: Name that the :class:`.Partition` instance will store.
    :param county_field_name: Name of county ID field on the graph.
    :return: The tracked data is a dictionary keyed on the county ID. The
              stored values are tuples of the form `(split, nodes, seen)`.
              `split` is a :class:`.CountySplit` enum, `nodes` is a list of
              node IDs, and `seen` is a list of assignment IDs that are
              contained in the county.
    """

    def _get_county_splits(partition):
        return compute_county_splits(partition, county_field_name, partition_name)

    return _get_county_splits


def compute_county_splits(partition, county_field, partition_field):
    """Track nodes in counties and information about their splitting."""

    # Create the initial county data containers.
    if not partition.parent:
        county_dict = dict()

        for node in partition.graph.node_indices:
            county = partition.graph.lookup(node, county_field)
            if county in county_dict:
                split, nodes, seen = county_dict[county]
            else:
                split, nodes, seen = CountySplit.NOT_SPLIT, [], set()

            nodes.append(node)
            seen.update(set([partition.assignment.mapping[node]]))

            if len(seen) > 1:
                split = CountySplit.OLD_SPLIT

            county_dict[county] = CountyInfo(split, nodes, seen)

        return county_dict

    new_county_dict = dict()

    parent = partition.parent
    for county, county_info in parent[partition_field].items():
        seen = set(partition.assignment.mapping[node] for node in county_info.nodes)

        split = CountySplit.NOT_SPLIT

        if len(seen) > 1:
            if county_info.split != CountySplit.OLD_SPLIT:
                split = CountySplit.NEW_SPLIT
            else:
                split = CountySplit.OLD_SPLIT

        new_county_dict[county] = CountyInfo(split, county_info.nodes, seen)

    return new_county_dict



from ._version import get_versions
from .chain import MarkovChain
from .graph import Graph
from .partition import GeographicPartition, Partition
from .updaters.election import Election

import geopandas

# warn about https://github.com/geopandas/geopandas/issues/2199
if geopandas.options.use_pygeos:
    raise ImportError(
        "GerryChain cannot use GeoPandas when PyGeos is enabled. Disable or "
        "uninstall PyGeos. You can disable PyGeos in GeoPandas by setting "
        "`geopandas.options.use_pygeos = False` before importing your shapefile."
    )

__version__ = get_versions()['version']
del get_versions



from itertools import product

from .constraints import Validator


def all_cut_edge_flips(partition):
    for edge, index in product(partition.cut_edges, (0, 1)):
        yield {edge[index]: partition.assignment.mapping[edge[1 - index]]}


def all_valid_states_one_flip_away(partition, constraints):
    """Generates all valid Partitions that differ from the given partition
    by one flip. These are the given partition's neighbors in the metagraph
    of partitions.
    """
    if callable(constraints):
        is_valid = constraints
    else:
        is_valid = Validator(constraints)

    for flip in all_cut_edge_flips(partition):
        next_state = partition.flip(flip)
        if is_valid(next_state):
            yield next_state


def all_valid_flips(partition, constraints):
    for state in all_valid_states_one_flip_away(partition, constraints):
        yield state.flips


def metagraph_degree(partition, constraints):
    return len(list(all_valid_states_one_flip_away(partition, constraints)))



from .constraints import Validator


class MarkovChain:
    """
    MarkovChain is an iterator that allows the user to iterate over the states
    of a Markov chain run.

    Example usage:

    .. code-block:: python

        chain = MarkovChain(proposal, constraints, accept, initial_state, total_steps)
        for state in chain:
            # Do whatever you want - print output, compute scores, ...

    """

    def __init__(self, proposal, constraints, accept, initial_state, total_steps):
        """
        :param proposal: Function proposing the next state from the current state.
        :param constraints: A function with signature ``Partition -> bool`` determining whether
            the proposed next state is valid (passes all binary constraints). Usually
            this is a :class:`~gerrychain.constraints.Validator` class instance.
        :param accept: Function accepting or rejecting the proposed state. In the most basic
            use case, this always returns ``True``. But if the user wanted to use a
            Metropolis-Hastings acceptance rule, this is where you would implement it.
        :param initial_state: Initial :class:`gerrychain.partition.Partition` class.
        :param total_steps: Number of steps to run.

        """
        if callable(constraints):
            is_valid = constraints
        else:
            is_valid = Validator(constraints)

        if not is_valid(initial_state):
            failed = [
                constraint
                for constraint in is_valid.constraints
                if not constraint(initial_state)
            ]
            message = (
                "The given initial_state is not valid according is_valid. "
                "The failed constraints were: " + ",".join([f.__name__ for f in failed])
            )
            raise ValueError(message)

        self.proposal = proposal
        self.is_valid = is_valid
        self.accept = accept
        self.total_steps = total_steps
        self.initial_state = initial_state
        self.state = initial_state

    def __iter__(self):
        self.counter = 0
        self.state = self.initial_state
        return self

    def __next__(self):
        if self.counter == 0:
            self.counter += 1
            return self.state

        while self.counter < self.total_steps:
            proposed_next_state = self.proposal(self.state)
            # Erase the parent of the parent, to avoid memory leak
            self.state.parent = None

            if self.is_valid(proposed_next_state):
                if self.accept(proposed_next_state):
                    self.state = proposed_next_state
                self.counter += 1
                return self.state
        raise StopIteration

    def __len__(self):
        return self.total_steps

    def __repr__(self):
        return "<MarkovChain [{} steps]>".format(len(self))

    def with_progress_bar(self):
        from tqdm.auto import tqdm

        return tqdm(self)



import json
import geopandas
import networkx

from gerrychain.graph.graph import FrozenGraph, Graph
from ..updaters import compute_edge_flows, flows_from_changes, cut_edges
from .assignment import get_assignment
from .subgraphs import SubgraphView


class Partition:
    """
    Partition represents a partition of the nodes of the graph. It will perform
    the first layer of computations at each step in the Markov chain - basic
    aggregations and calculations that we want to optimize.

    :ivar gerrychain.Graph graph: The underlying graph.
    :ivar gerrychain.Assignment assignment: Maps node IDs to district IDs.
    :ivar dict parts: Maps district IDs to the set of nodes in that district.
    :ivar dict subgraphs: Maps district IDs to the induced subgraph of that district.
    """
    __slots__ = (
        'graph',
        'subgraphs',
        'assignment',
        'updaters',
        'parent',
        'flips',
        'flows',
        'edge_flows',
        '_cache'
    )

    default_updaters = {"cut_edges": cut_edges}

    def __init__(
        self, graph=None, assignment=None, updaters=None, parent=None, flips=None,
        use_default_updaters=True
    ):
        """
        :param graph: Underlying graph.
        :param assignment: Dictionary assigning nodes to districts.
        :param updaters: Dictionary of functions to track data about the partition.
            The keys are stored as attributes on the partition class,
            which the functions compute.
        :param use_default_updaters: If `False`, do not include default updaters.
        """
        if parent is None:
            self._first_time(graph, assignment, updaters, use_default_updaters)
        else:
            self._from_parent(parent, flips)

        self._cache = dict()
        self.subgraphs = SubgraphView(self.graph, self.parts)

    def _first_time(self, graph, assignment, updaters, use_default_updaters):
        if isinstance(graph, Graph):
            self.graph = FrozenGraph(graph)
        elif isinstance(graph, networkx.Graph):
            graph = Graph.from_networkx(graph)
            self.graph = FrozenGraph(graph)
        elif isinstance(graph, FrozenGraph):
            self.graph = graph
        else:
            raise TypeError("Unsupported Graph object")

        self.assignment = get_assignment(assignment, graph)

        if set(self.assignment) != set(graph):
            raise KeyError("The graph's node labels do not match the Assignment's keys")

        if updaters is None:
            updaters = dict()

        if use_default_updaters:
            self.updaters = self.default_updaters
        else:
            self.updaters = {}

        self.updaters.update(updaters)

        self.parent = None
        self.flips = None
        self.flows = None
        self.edge_flows = None

    def _from_parent(self, parent, flips):
        self.parent = parent
        self.flips = flips

        self.graph = parent.graph
        self.updaters = parent.updaters

        self.flows = flows_from_changes(parent, self)  # careful

        self.assignment = parent.assignment.copy()
        self.assignment.update_flows(self.flows)

        if "cut_edges" in self.updaters:
            self.edge_flows = compute_edge_flows(self)

    def __repr__(self):
        number_of_parts = len(self)
        s = "s" if number_of_parts > 1 else ""
        return "<{} [{} part{}]>".format(self.__class__.__name__, number_of_parts, s)

    def __len__(self):
        return len(self.parts)

    def flip(self, flips):
        """Returns the new partition obtained by performing the given `flips`
        on this partition.

        :param flips: dictionary assigning nodes of the graph to their new districts
        :return: the new :class:`Partition`
        :rtype: Partition
        """
        return self.__class__(parent=self, flips=flips)

    def crosses_parts(self, edge):
        """Answers the question "Does this edge cross from one part of the
        partition to another?

        :param edge: tuple of node IDs
        :rtype: bool
        """
        return self.assignment.mapping[edge[0]] != self.assignment.mapping[edge[1]]

    def __getitem__(self, key):
        """Allows accessing the values of updaters computed for this
        Partition instance.

        :param key: Property to access.
        """
        if key not in self._cache:
            self._cache[key] = self.updaters[key](self)
        return self._cache[key]

    def __getattr__(self, key):
        return self[key]

    def keys(self):
        return self.updaters.keys()

    @property
    def parts(self):
        return self.assignment.parts

    def plot(self, geometries=None, **kwargs):
        """Plot the partition, using the provided geometries.

        :param geometries: A :class:`geopandas.GeoDataFrame` or :class:`geopandas.GeoSeries`
            holding the geometries to use for plotting. Its :class:`~pandas.Index` should match
            the node labels of the partition's underlying :class:`~gerrychain.Graph`.
        :param `**kwargs`: Additional arguments to pass to :meth:`geopandas.GeoDataFrame.plot`
            to adjust the plot.
        """
        if geometries is None:
            geometries = self.graph.geometry

        if set(geometries.index) != set(self.graph.nodes):
            raise TypeError(
                "The provided geometries do not match the nodes of the graph."
            )
        assignment_series = self.assignment.to_series()
        if isinstance(geometries, geopandas.GeoDataFrame):
            geometries = geometries.geometry
        df = geopandas.GeoDataFrame(
            {"assignment": assignment_series}, geometry=geometries
        )
        return df.plot(column="assignment", **kwargs)

    @classmethod
    def from_districtr_file(cls, graph, districtr_file, updaters=None):
        """Create a Partition from a districting plan created with `Districtr`_,
        a free and open-source web app created by MGGG for drawing districts.

        The provided ``graph`` should be created from the same shapefile as the
        Districtr module used to draw the districting plan. These shapefiles may
        be found in a repository in the `mggg-states`_ GitHub organization, or by
        request from MGGG.

        .. _`Districtr`: https://mggg.org/Districtr
        .. _`mggg-states`: https://github.com/mggg-states

        :param graph: :class:`~gerrychain.Graph`
        :param districtr_file: the path to the ``.json`` file exported from Districtr
        :param updaters: dictionary of updaters
        """
        with open(districtr_file) as f:
            districtr_plan = json.load(f)

        id_column_key = districtr_plan["idColumn"]["key"]
        districtr_assignment = districtr_plan["assignment"]
        try:
            node_to_id = {node: str(graph.nodes[node][id_column_key]) for node in graph}
        except KeyError:
            raise TypeError(
                "The provided graph is missing the {} column, which is "
                "needed to match the Districtr assignment to the nodes of the graph."
            )

        assignment = {node: districtr_assignment[node_to_id[node]] for node in graph}

        return cls(graph, assignment, updaters)



# Imports
from collections import defaultdict, Counter
import networkx as nx
import math


class LocalitySplits:

    """Computes various splitting measures for a partition


    Can be used to compute how a districting plan splits
    against any static attribute. The prototypical example
    is to consider how a districting plan subdivides counties
    or municipalities, but other units, such as city
    neighborhoods, state legislative districts, or Census
    tracts could be treated as 'localities'

    Example usage::
        # Assuming your nodes have attributes "countyID"
        # with (for example) the name of the county that
        # node lies in and a population attribute "pop":
        county_splits = LocalitySplits(
            "countysplits",
            "countyID",
            "pop",
            ["num_parts", "symmetric_entropy","power_entropy"],
            pent_alpha = 0.8
        )
        # Assuming you already have a graph and assignment:
        partition = Partition(
            graph,
            assignment,
            updaters={"county_splits" : county_splits}
        )
        # The updater returns an dictionary instance, which
        # at each step of the chain has the name of the score
        # and its value at that step

    """

    def __init__(self, name, col_id, pop_col,
                 scores_to_compute=["num_parts"], pent_alpha=.05):
        """
        :param name: The name of the updater (e.g. "countysplits")
        :param col_id: The name of the column containing the locality
            attribute (i.e. county ids, municipality names, etc.)
        :param pop_col: The name of the column containing population counts.
        :param scores_to_compute: T list/tuple/set of strings naming the
            score functions to compute at each step. This should be
            some subcollection of ```['num_parts', 'num_pieces',
            'naked_boundary', 'shannon_entropy', 'power_entropy',
            'symmetric_entropy', 'num_split_localities']```
        :param pent_alpha: A number between 0 and 1 which is
            passed as the exponent to :meth:`~LocalitySplits.power_entropy`
        """

        self.name = name
        self.col_id = col_id

        self.pop_col = pop_col

        self.pent_alpha = pent_alpha

        self.localities = []
        self.localitydict = {}
        self.locality_splits = {}
        self.locality_splits_inv = {}

        # A dictionary containing the number minimum number
        # of districts which a locality must touch. I.e. if
        # the ideal district population is 10,000 and a
        # locality has 35,000 people, then that locality
        # must be in at least four districts.  Not
        # presently used to compute any score functions,
        # but may be useful for future development or
        # certain use cases.
        self.allowed_pieces = {}

        self.scores = dict.fromkeys(scores_to_compute)

    def __call__(self, partition):
        if self.localities == []:

            self.localitydict = dict(partition.graph.nodes(data=self.col_id))
            self.localities = set(list(self.localitydict.values()))

        locality_splits = {k: [self.localitydict[v] for v in d]
                           for k, d in partition.assignment.parts.items()}
        self.locality_splits = {k: Counter(v)
                                for k, v in locality_splits.items()}

        self.locality_splits_inv = defaultdict(dict)
        for k, v in self.locality_splits.items():
            for k2, v2 in v.items():
                self.locality_splits_inv[k2][k] = v2

        if self.allowed_pieces == {}:

            allowed_pieces = {}

            totpop = 0
            for node in partition.graph.nodes:
                totpop += partition.graph.nodes[node][self.pop_col]

            num_districts = len(partition.assignment.parts.keys())

            for loc in self.localities:
                sg = partition.graph.subgraph(
                    n for n, v in partition.graph.nodes(
                        data=True) if v[self.col_id] == loc)

                pop = 0
                for n in sg.nodes():
                    pop += sg.nodes[n][self.pop_col]

                allowed_pieces[loc] = math.ceil(pop / (totpop / num_districts))
            self.allowed_pieces = allowed_pieces

        for s in self.scores:
            if s == 'num_parts':
                self.scores[s] = self.num_parts(partition)

            if s == 'num_pieces':
                self.scores[s] = self.num_pieces(partition)

            if s == 'naked_boundary':
                self.scores[s] = self.naked_boundary(partition)

            if s == 'shannon_entropy':
                self.scores[s] = self.shannon_entropy(partition)

            if s == 'power_entropy':
                self.scores[s] = self.power_entropy(partition)

            if s == 'symmetric_entropy':
                self.scores[s] = self.symmetric_entropy(partition)

            if s == 'num_split_localities':
                self.scores[s] = self.num_split_localities(partition)

        return self.scores

    def num_parts(self, partition):
        '''
        Calculates the number of unique locality-district pairs.

        :param partition: The partition to be scored.

        :return: The number of parts, i.e. the number of unique
            locality-district pairs.
        '''

        counter = 0
        for district in self.locality_splits.keys():
            counter += len(self.locality_splits[district])
        return counter

    def num_pieces(self, partition):
        '''
        Calculates the number of pieces.

        :param partition: The partition to be scored.

        :return: Number of pieces, where each piece is formed by
            cutting the graph by both locality and district boundaries.

        '''
        locality_intersections = {}

        for n in partition.graph.nodes():
            locality = partition.graph.nodes[n][self.col_id]
            if locality not in locality_intersections:
                locality_intersections[locality] = set(
                    [partition.assignment.mapping[n]])

            locality_intersections[locality].update([partition.assignment.mapping[n]])

        pieces = 0
        for locality in locality_intersections:
            for d in locality_intersections[locality]:
                subgraph = partition.graph.subgraph(
                    [x for x in partition.parts[d]
                        if partition.graph.nodes[x][self.col_id] == locality]
                )

                pieces += nx.number_connected_components(subgraph)
        return pieces

    def naked_boundary(self, partition):
        '''
        Computes the number of cut edges inside localities (i.e. the
            number of cut edges with both endpoints in the same locality).

        :param partition: The partition to be scored.

        :return: The number of cut edges within a locality.
        '''

        cut_edges_within = 0
        cut_edge_set = partition["cut_edges"]
        for i in cut_edge_set:
            vtd_1 = i[0]
            vtd_2 = i[1]
            county_1 = self.localitydict.get(vtd_1)
            county_2 = self.localitydict.get(vtd_2)
            if county_1 == county_2:  # not on county boundary
                cut_edges_within += 1
        return cut_edges_within

    def shannon_entropy(self, partition):
        '''
        Computes the shannon entropy score of a district plan.

        :param partition: The partition to be scored.

        :returns: Shannon entropy score.
        '''

        total_vtds = 0
        for k, v in self.locality_splits.items():
            for x in list(v.values()):
                total_vtds += x

        entropy = 0
        for locality_j in self.localities:  # iter thru locs to get total count
            tot_county_vtds = 0
            # iter thru counters
            for k, v in self.locality_splits.items():
                v = dict(v)
                if locality_j in list(v.keys()):
                    tot_county_vtds += v[locality_j]

            inner_sum = 0
            q = tot_county_vtds / total_vtds

            # iter thru districts to get vtds in county in district
            # for district in range(num_districts):
            for k, v in self.locality_splits.items():
                # counter = dict(locality_splits[district+1])
                count = dict(v)
                if locality_j in count:
                    intersection = count[str(locality_j)]
                    p = intersection / tot_county_vtds

                    if p != 0:
                        inner_sum += p * math.log(1 / p)

            entropy += q * (inner_sum)
        return entropy

    def power_entropy(self, partition):

        '''
        Computes the power entropy score of a district plan.

        :param partition: The partition to be scored.

        :returns: Power entropy score.
        '''

        total_vtds = 0  # count the total number of vtds in state
        for k, v in self.locality_splits.items():
            for x in list(v.values()):
                total_vtds += x

        entropy = 0
        for locality_j in self.localities:  # iter thru locs to get total count
            tot_county_vtds = 0
            # iter thru counters
            for k, v in self.locality_splits.items():
                v = dict(v)
                if locality_j in list(v.keys()):
                    tot_county_vtds += v[locality_j]

            inner_sum = 0

            q = tot_county_vtds / total_vtds
            # iter thru districts to get vtds in county in district
            # for district in range(num_districts):
            for k, v in self.locality_splits.items():
                # counter = dict(locality_splits[district+1])
                count = dict(v)
                if locality_j in count:
                    intersection = count[str(locality_j)]
                    p = intersection / tot_county_vtds

                    if p != 0:
                        inner_sum += p ** (1 - self.pent_alpha)

            entropy += 1 / q * (inner_sum - 1)
        return entropy

    def symmetric_entropy(self, partition):  # IN PROGRESS
        '''
        Calculates the symmetric entropy score.

        :param partition: The partition to be scored.

        :return: The symmetric square root entropy score.
        '''

        district_dict = dict(partition.parts)

        for district in district_dict.keys():
            vtds = district_dict[district]
            locality_pop = {k: 0 for k in self.localities}
            for vtd in vtds:
                locality_pop[self.localitydict[vtd]] += partition.graph.nodes[
                    vtd][self.pop_col]
            district_dict[district] = locality_pop

        district_dict_inv = defaultdict(dict)
        for k, v in district_dict.items():
            for k2, v2 in v.items():
                district_dict_inv[k2][k] = v2

        # how do districts split localities?
        score = 0
        for district in district_dict.keys():
            localities_and_pops = district_dict[district]
            total = sum(localities_and_pops.values())
            fractional_sum = 0
            for locality in localities_and_pops.keys():
                fractional_sum += math.sqrt(
                    localities_and_pops[locality] / total)
            score += total * fractional_sum

        # how do localities split districts?
        for locality in district_dict_inv.keys():
            districts_and_pops = district_dict_inv[locality]
            total = sum(districts_and_pops.values())
            fractional_sum = 0
            for district in districts_and_pops.keys():
                fractional_sum += math.sqrt(districts_and_pops[district]
                                            / total)
            score += total * fractional_sum

        return score

    def num_split_localities(self, partition):
        '''
        Calculates the number of localities touching 2 or more districts.

        :param partition: The partition to be scored.

        :return: The number of split localities, i.e. the number of localities
            touching 2 or more districts.
        '''

        total_splits = 0

        for v in self.locality_splits_inv.values():
            if len(v) > 1:
                total_splits += 1

        return total_splits



from gerrychain.partition import Partition
from gerrychain.updaters import (
    Tally,
    boundary_nodes,
    cut_edges,
    cut_edges_by_part,
    exterior_boundaries,
    interior_boundaries,
    perimeter,
)


class GeographicPartition(Partition):
    """A :class:`Partition` with areas, perimeters, and boundary information included.
    These additional data allow you to compute compactness scores like
    `Polsby-Popper <https://en.wikipedia.org/wiki/Polsby-Popper_Test>`_.
    """

    default_updaters = {
        "perimeter": perimeter,
        "exterior_boundaries": exterior_boundaries,
        "interior_boundaries": interior_boundaries,
        "boundary_nodes": boundary_nodes,
        "cut_edges": cut_edges,
        "area": Tally("area", alias="area"),
        "cut_edges_by_part": cut_edges_by_part,
    }



import networkx as nx
from networkx.algorithms import tree

from functools import partial
from .random import random
from collections import deque, namedtuple


def predecessors(h, root):
    return {a: b for a, b in nx.bfs_predecessors(h, root)}


def successors(h, root):
    return {a: b for a, b in nx.bfs_successors(h, root)}


def random_spanning_tree(graph):
    """ Builds a spanning tree chosen by Kruskal's method using random weights.
        :param graph: FrozenGraph

        Important Note:
        The key is specifically labelled "random_weight" instead of the previously
        used "weight". Turns out that networkx uses the "weight" keyword for other
        operations, like when computing the laplacian or the adjacency matrix.
        This meant that the laplacian would change for the graph step to step,
        something that we do not intend!!
    """
    for edge in graph.edge_indices:
        graph.edges[edge]["random_weight"] = random.random()

    spanning_tree = tree.maximum_spanning_tree(
        graph, algorithm="kruskal", weight="random_weight"
    )
    return spanning_tree


def uniform_spanning_tree(graph, choice=random.choice):
    """ Builds a spanning tree chosen uniformly from the space of all
        spanning trees of the graph.
        :param graph: Networkx Graph
        :param choice: :func:`random.choice`
    """
    root = choice(graph.node_indices)
    tree_nodes = set([root])
    next_node = {root: None}

    for node in graph.node_indices:
        u = node
        while u not in tree_nodes:
            next_node[u] = choice(graph.neighbors(u))
            u = next_node[u]

        u = node
        while u not in tree_nodes:
            tree_nodes.add(u)
            u = next_node[u]

    G = nx.Graph()
    for node in tree_nodes:
        if next_node[node] is not None:
            G.add_edge(node, next_node[node])

    return G


class PopulatedGraph:
    def __init__(self, graph, populations, ideal_pop, epsilon):
        self.graph = graph
        self.subsets = {node: {node} for node in graph.node_indices}
        self.population = populations.copy()
        self.tot_pop = sum(self.population.values())
        self.ideal_pop = ideal_pop
        self.epsilon = epsilon
        self._degrees = {node: graph.degree(node) for node in graph.node_indices}

    def __iter__(self):
        return iter(self.graph)

    def degree(self, node):
        return self._degrees[node]

    def contract_node(self, node, parent):
        self.population[parent] += self.population[node]
        self.subsets[parent] |= self.subsets[node]
        self._degrees[parent] -= 1

    def has_ideal_population(self, node):
        return (
            abs(self.population[node] - self.ideal_pop) < self.epsilon * self.ideal_pop
        )


Cut = namedtuple("Cut", "edge subset")


def find_balanced_edge_cuts_contraction(h, choice=random.choice):
    # this used to be greater than 2 but failed on small grids:(
    root = choice([x for x in h if h.degree(x) > 1])
    # BFS predecessors for iteratively contracting leaves
    pred = predecessors(h.graph, root)

    cuts = []
    leaves = deque(x for x in h if h.degree(x) == 1)
    while len(leaves) > 0:
        leaf = leaves.popleft()
        if h.has_ideal_population(leaf):
            cuts.append(Cut(edge=(leaf, pred[leaf]), subset=h.subsets[leaf].copy()))
        # Contract the leaf:
        parent = pred[leaf]
        h.contract_node(leaf, parent)
        if h.degree(parent) == 1 and parent != root:
            leaves.append(parent)
    return cuts


def find_balanced_edge_cuts_memoization(h, choice=random.choice):
    root = choice([x for x in h if h.degree(x) > 1])
    pred = predecessors(h.graph, root)
    succ = successors(h.graph, root)
    total_pop = h.tot_pop
    subtree_pops = {}
    stack = deque(n for n in succ[root])
    while stack:
        next_node = stack.pop()
        if next_node not in subtree_pops:
            if next_node in succ:
                children = succ[next_node]
                if all(c in subtree_pops for c in children):
                    subtree_pops[next_node] = sum(subtree_pops[c] for c in children)
                    subtree_pops[next_node] += h.population[next_node]
                else:
                    stack.append(next_node)
                    for c in children:
                        if c not in subtree_pops:
                            stack.append(c)
            else:
                subtree_pops[next_node] = h.population[next_node]

    cuts = []
    for node, tree_pop in subtree_pops.items():

        def part_nodes(start):
            nodes = set()
            queue = deque([start])
            while queue:
                next_node = queue.pop()
                if next_node not in nodes:
                    nodes.add(next_node)
                    if next_node in succ:
                        for c in succ[next_node]:
                            if c not in nodes:
                                queue.append(c)
            return nodes

        if abs(tree_pop - h.ideal_pop) <= h.ideal_pop * h.epsilon:
            cuts.append(Cut(edge=(node, pred[node]), subset=part_nodes(node)))
        elif abs((total_pop - tree_pop) - h.ideal_pop) <= h.ideal_pop * h.epsilon:
            cuts.append(Cut(edge=(node, pred[node]),
                            subset=set(h.graph.nodes) - part_nodes(node)))
    return cuts


def bipartition_tree(
    graph,
    pop_col,
    pop_target,
    epsilon,
    node_repeats=1,
    spanning_tree=None,
    spanning_tree_fn=random_spanning_tree,
    balance_edge_fn=find_balanced_edge_cuts_memoization,
    choice=random.choice,
    max_attempts=None
):
    """This function finds a balanced 2 partition of a graph by drawing a
    spanning tree and finding an edge to cut that leaves at most an epsilon
    imbalance between the populations of the parts. If a root fails, new roots
    are tried until node_repeats in which case a new tree is drawn.

    Builds up a connected subgraph with a connected complement whose population
    is ``epsilon * pop_target`` away from ``pop_target``.

    Returns a subset of nodes of ``graph`` (whose induced subgraph is connected).
    The other part of the partition is the complement of this subset.

    :param graph: The graph to partition
    :param pop_col: The node attribute holding the population of each node
    :param pop_target: The target population for the returned subset of nodes
    :param epsilon: The allowable deviation from  ``pop_target`` (as a percentage of
        ``pop_target``) for the subgraph's population
    :param node_repeats: A parameter for the algorithm: how many different choices
        of root to use before drawing a new spanning tree.
    :param spanning_tree: The spanning tree for the algorithm to use (used when the
        algorithm chooses a new root and for testing)
    :param spanning_tree_fn: The random spanning tree algorithm to use if a spanning
        tree is not provided
    :param choice: :func:`random.choice`. Can be substituted for testing.
    :param max_atempts: The max number of attempts that should be made to bipartition.
    """
    populations = {node: graph.nodes[node][pop_col] for node in graph.node_indices}

    possible_cuts = []
    if spanning_tree is None:
        spanning_tree = spanning_tree_fn(graph)

    restarts = 0
    attempts = 0
    while max_attempts is None or attempts < max_attempts:
        if restarts == node_repeats:
            spanning_tree = spanning_tree_fn(graph)
            restarts = 0
        h = PopulatedGraph(spanning_tree, populations, pop_target, epsilon)
        possible_cuts = balance_edge_fn(h, choice=choice)

        if len(possible_cuts) != 0:
            return choice(possible_cuts).subset

        restarts += 1
        attempts += 1

    raise RuntimeError(f"Could not find a possible cut after {max_attempts} attempts.")


def _bipartition_tree_random_all(
    graph,
    pop_col,
    pop_target,
    epsilon,
    node_repeats=1,
    repeat_until_valid=True,
    spanning_tree=None,
    spanning_tree_fn=random_spanning_tree,
    balance_edge_fn=find_balanced_edge_cuts_memoization,
    choice=random.choice,
    max_attempts=None
):
    """Randomly bipartitions a graph and returns all cuts."""
    populations = {node: graph.nodes[node][pop_col] for node in graph.node_indices}

    possible_cuts = []
    if spanning_tree is None:
        spanning_tree = spanning_tree_fn(graph)

    repeat = True
    attempts = 0
    while max_attempts is None or attempts < max_attempts:
        spanning_tree = spanning_tree_fn(graph)
        h = PopulatedGraph(spanning_tree, populations, pop_target, epsilon)
        possible_cuts = balance_edge_fn(h, choice=choice)

        repeat = repeat_until_valid
        attempts += 1

        if not (repeat and len(possible_cuts) == 0):
            return possible_cuts

    raise RuntimeError(f"Could not find a possible cut after {max_attempts} attempts.")


def bipartition_tree_random(
    graph,
    pop_col,
    pop_target,
    epsilon,
    node_repeats=1,
    repeat_until_valid=True,
    spanning_tree=None,
    spanning_tree_fn=random_spanning_tree,
    balance_edge_fn=find_balanced_edge_cuts_memoization,
    choice=random.choice,
    *args, **kwargs
):
    """This is like :func:`bipartition_tree` except it chooses a random balanced
    cut, rather than the first cut it finds.

    This function finds a balanced 2 partition of a graph by drawing a
    spanning tree and finding an edge to cut that leaves at most an epsilon
    imbalance between the populations of the parts. If a root fails, new roots
    are tried until node_repeats in which case a new tree is drawn.

    Builds up a connected subgraph with a connected complement whose population
    is ``epsilon * pop_target`` away from ``pop_target``.

    Returns a subset of nodes of ``graph`` (whose induced subgraph is connected).
    The other part of the partition is the complement of this subset.

    :param graph: The graph to partition
    :param pop_col: The node attribute holding the population of each node
    :param pop_target: The target population for the returned subset of nodes
    :param epsilon: The allowable deviation from  ``pop_target`` (as a percentage of
        ``pop_target``) for the subgraph's population
    :param node_repeats: A parameter for the algorithm: how many different choices
        of root to use before drawing a new spanning tree.
    :param repeat_until_valid: Determines whether to keep drawing spanning trees
        until a tree with a balanced cut is found. If `True`, a set of nodes will
        always be returned; if `False`, `None` will be returned if a valid spanning
        tree is not found on the first try.
    :param spanning_tree: The spanning tree for the algorithm to use (used when the
        algorithm chooses a new root and for testing)
    :param spanning_tree_fn: The random spanning tree algorithm to use if a spanning
        tree is not provided
    :param balance_edge_fn: The algorithm used to find balanced cut edges
    :param choice: :func:`random.choice`. Can be substituted for testing.
    """
    possible_cuts = _bipartition_tree_random_all(*args, **kwargs)
    if possible_cuts:
        return choice(possible_cuts).subset


def recursive_tree_part(
    graph,
    parts,
    pop_target,
    pop_col,
    epsilon,
    node_repeats=1,
    method=partial(bipartition_tree, max_attempts=10000)
):
    """Uses :func:`~gerrychain.tree.bipartition_tree` recursively to partition a tree into
    ``len(parts)`` parts of population ``pop_target`` (within ``epsilon``). Can be used to
    generate initial seed plans or to implement ReCom-like "merge walk" proposals.

    :param graph: The graph
    :param parts: Iterable of part labels (like ``[0,1,2]`` or ``range(4)``
    :param pop_target: Target population for each part of the partition
    :param pop_col: Node attribute key holding population data
    :param epsilon: How far (as a percentage of ``pop_target``) from ``pop_target`` the parts
        of the partition can be
    :param node_repeats: Parameter for :func:`~gerrychain.tree_methods.bipartition_tree` to use.
    :param method: The partition method to use.
    :return: New assignments for the nodes of ``graph``.
    :rtype: dict
    """
    flips = {}
    remaining_nodes = graph.node_indices
    # We keep a running tally of deviation from ``epsilon`` at each partition
    # and use it to tighten the population constraints on a per-partition
    # basis such that every partition, including the last partition, has a
    # population within +/-``epsilon`` of the target population.
    # For instance, if district n's population exceeds the target by 2%
    # with a +/-2% epsilon, then district n+1's population should be between
    # 98% of the target population and the target population.
    debt = 0

    for part in parts[:-1]:
        min_pop = max(pop_target * (1 - epsilon), pop_target * (1 - epsilon) - debt)
        max_pop = min(pop_target * (1 + epsilon), pop_target * (1 + epsilon) - debt)
        nodes = method(
            graph.subgraph(remaining_nodes),
            pop_col=pop_col,
            pop_target=(min_pop + max_pop) / 2,
            epsilon=(max_pop - min_pop) / (2 * pop_target),
            node_repeats=node_repeats,
        )

        if nodes is None:
            raise BalanceError()

        part_pop = 0
        for node in nodes:
            flips[node] = part
            part_pop += graph.nodes[node][pop_col]
        debt += part_pop - pop_target
        remaining_nodes -= nodes

    # All of the remaining nodes go in the last part
    for node in remaining_nodes:
        flips[node] = parts[-1]

    return flips


def get_seed_chunks(
    graph,
    num_chunks,
    num_dists,
    pop_target,
    pop_col,
    epsilon,
    node_repeats=1,
    method=partial(bipartition_tree_random, max_attempts=10000)
):
    """
    Helper function for recursive_seed_part. Partitions the graph into ``num_chunks`` chunks,
    balanced within new_epsilon <= ``epsilon`` of a balanced target population.

    :param graph: The graph
    :param parts: Iterable of part labels (like ``[0,1,2]`` or ``range(4)``
    :param pop_target: target population of the districts (not of the chunks)
    :param pop_col: Node attribute key holding population data
    :param epsilon: How far (as a percentage of ``pop_target``) from ``pop_target`` the parts
        of the partition can be
    :param node_repeats: Parameter for :func:`~gerrychain.tree_methods.bipartition_tree` to use.
    :return: New assignments for the nodes of ``graph``.
    :rtype: dict
    """
    num_chunks_left = num_dists // num_chunks
    parts = range(num_chunks)
    new_epsilon = epsilon / (num_chunks_left * num_chunks)
    if num_chunks_left == 1:
        new_epsilon = epsilon

    chunk_pop = 0
    for node in graph.node_indices:
        chunk_pop += graph.nodes[node][pop_col]

    while True:
        epsilon = abs(epsilon)

        flips = {}
        remaining_nodes = set(graph.nodes)

        min_pop = pop_target * (1 - new_epsilon) * num_chunks_left
        max_pop = pop_target * (1 + new_epsilon) * num_chunks_left

        chunk_pop_target = chunk_pop / num_chunks

        diff = min(max_pop - chunk_pop_target, chunk_pop_target - min_pop)
        new_new_epsilon = diff / chunk_pop_target

        for i in range(len(parts[:-1])):
            part = parts[i]

            nodes = method(
                graph.subgraph(remaining_nodes),
                pop_col=pop_col,
                pop_target=chunk_pop_target,
                epsilon=new_new_epsilon,
                node_repeats=node_repeats,
            )

            if nodes is None:
                raise BalanceError()

            for node in nodes:
                flips[node] = part
            remaining_nodes -= nodes

            # All of the remaining nodes go in the last part
            for node in remaining_nodes:
                flips[node] = parts[-1]

        part_pop = 0
        for node in remaining_nodes:
            part_pop += graph.nodes[node][pop_col]
        part_pop_as_dist = part_pop / num_chunks_left
        fake_epsilon = epsilon
        if num_chunks_left != 1:
            fake_epsilon = epsilon / 2
        min_pop_as_dist = pop_target * (1 - fake_epsilon)
        max_pop_as_dist = pop_target * (1 + fake_epsilon)

        if part_pop_as_dist < min_pop_as_dist:
            new_epsilon = new_epsilon / 2
        elif part_pop_as_dist > max_pop_as_dist:
            new_epsilon = new_epsilon / 2
        else:
            break

    chunks = {}
    for key in flips.keys():
        if flips[key] not in chunks.keys():
            chunks[flips[key]] = []
        chunks[flips[key]].append(key)

    return list(chunks.values())


def get_max_prime_factor_less_than(
    n, ceil
):
    """
    Helper function for recursive_seed_part. Returns the largest prime factor of ``n`` less than
    ``ceil``, or None if all are greater than ceil.
    """
    factors = []
    i = 2
    while i * i <= n:
        if n % i:
            i += 1
        else:
            n //= i
            factors.append(i)
    if n > 1:
        factors.append(n)

    if len(factors) == 0:
        return 1
    m = [i for i in factors if i <= ceil]
    if m == []:
        return None
    return int(max(m))


def recursive_seed_part_inner(
    graph,
    num_dists,
    pop_target,
    pop_col,
    epsilon,
    method=partial(bipartition_tree, max_attempts=10000),
    node_repeats=1,
    n=None,
    ceil=None,
):
    """
    Inner function for recursive_seed_part.
    Returns a partition with ``num_dists`` districts balanced within ``epsilon`` of
    ``pop_target``.
    Splits graph into num_chunks chunks, and then recursively splits each chunk into
    ``num_dists``/num_chunks chunks.
    The number num_chunks of chunks is chosen based on ``n`` and ``ceil`` as follows:
        If ``n`` is None, and ``ceil`` is None, num_chunks is the largest prime factor
        of ``num_dists``.
        If ``n`` is None and ``ceil`` is an integer at least 2, then num_chunks is the
        largest prime factor of ``num_dists`` that is less than ``ceil``
        If ``n`` is a positive integer, num_chunks equals n.
    Finally, if the number of chunks as chosen above does not divide ``num_dists``, then
    this function bites off a single district from the graph and recursively partitions
    the remaining graph into ``num_dists``-1 districts.

    :param graph: The graph
    :param num_dists: number of districts to partition the graph into
    :param pop_target: Target population for each part of the partition
    :param pop_col: Node attribute key holding population data
    :param epsilon: How far (as a percentage of ``pop_target``) from ``pop_target`` the parts
        of the partition can be
    :param method: Function used to find balanced partitions at the 2-district level
    :param node_repeats: Parameter for :func:`~gerrychain.tree_methods.bipartition_tree` to use.
    :param n: Either a positive integer (greater than 1) or None. If n is a positive integer,
        this function will recursively create a seed plan by either biting off districts from
        graph or dividing graph into n chunks and recursing into each of these. If n is None,
        this function prime factors ``num_dists``=n_1*n_2*...*n_k (n_1 > n_2 > ... n_k) and
        recursively partitions graph into n_1 chunks.
    :param ceil: Either a positive integer (at least 2) or None. Relevant only if n is None.
        If ``ceil`` is a positive integer then finds the largest factor of ``num_dists`` less
        than or equal to ``ceil``, and recursively splits graph into that number of chunks, or
        bites off a district if that number is 1.
    :return: New assignments for the nodes of ``graph``.
    :rtype: List of lists, each list is a district
    """
    # Chooses num_chunks
    if n is None:
        if ceil is None:
            num_chunks = get_max_prime_factor_less_than(num_dists, num_dists)
        elif ceil >= 2:
            num_chunks = get_max_prime_factor_less_than(num_dists, ceil)
        else:
            raise ValueError("ceil must be None or at least 2")
    elif n > 1:
        num_chunks = n
    else:
        raise ValueError("n must be None or a positive integer")

    # base case
    if num_dists == 1:
        return [set(graph.nodes)]

    # bite off a district and recurse into the remaining subgraph
    elif num_chunks is None or num_dists % num_chunks != 0:
        remaining_nodes = set(graph.nodes)
        nodes = method(
            graph.subgraph(remaining_nodes),
            pop_col=pop_col,
            pop_target=pop_target,
            epsilon=epsilon,
            node_repeats=node_repeats
        )
        remaining_nodes -= nodes
        assignment = [nodes] + recursive_seed_part_inner(graph.subgraph(remaining_nodes),
            num_dists - 1,
            pop_target,
            pop_col,
            epsilon,
            n=n,
            ceil=ceil)

    # split graph into num_chunks chunks, and recurse into each chunk
    elif num_dists % num_chunks == 0:
        chunks = get_seed_chunks(
            graph,
            num_chunks,
            num_dists,
            pop_target,
            pop_col,
            epsilon,
            method=method
        )

        assignment = []
        for chunk in chunks:
            chunk_assignment = recursive_seed_part_inner(
                graph.subgraph(chunk),
                num_dists // num_chunks,
                pop_target,
                pop_col,
                epsilon,
                n=n,
                ceil=ceil
            )
            assignment += chunk_assignment

    return assignment


def recursive_seed_part(
    graph,
    parts,
    pop_target,
    pop_col,
    epsilon,
    method=partial(bipartition_tree, max_attempts=10000),
    node_repeats=1,
    n=None,
    ceil=None
):
    """
    Returns a partition with ``num_dists`` districts balanced within ``epsilon`` of
    ``pop_target`` by recursively splitting graph using recursive_seed_part_inner.

    :param graph: The graph
    :param parts: Iterable of part labels (like ``[0,1,2]`` or ``range(4)``
    :param pop_target: Target population for each part of the partition
    :param pop_col: Node attribute key holding population data
    :param epsilon: How far (as a percentage of ``pop_target``) from ``pop_target`` the parts
        of the partition can be
    :param method: Function used to find balanced partitions at the 2-district level
    :param node_repeats: Parameter for :func:`~gerrychain.tree_methods.bipartition_tree` to use.
    :param n: Either a positive integer (greater than 1) or None. If n is a positive integer,
    this function will recursively create a seed plan by either biting off districts from graph
    or dividing graph into n chunks and recursing into each of these. If n is None, this
    function prime factors ``num_dists``=n_1*n_2*...*n_k (n_1 > n_2 > ... n_k) and recursively
    partitions graph into n_1 chunks.
    :param ceil: Either a positive integer (at least 2) or None. Relevant only if n is None. If
    ``ceil`` is a positive integer then finds the largest factor of ``num_dists`` less than or
    equal to ``ceil``, and recursively splits graph into that number of chunks, or bites off a
    district if that number is 1.
    :return: New assignments for the nodes of ``graph``.
    :rtype: dict
    """
    flips = {}
    assignment = recursive_seed_part_inner(
        graph,
        len(parts),
        pop_target,
        pop_col,
        epsilon,
        method=partial(bipartition_tree, max_attempts=10000),
        node_repeats=node_repeats,
        n=n,
        ceil=ceil
    )
    for i in range(len(assignment)):
        for node in assignment[i]:
            flips[node] = parts[i]
    return flips


class BalanceError(Exception):
    """Raised when a balanced cut cannot be found."""



from .compactness import (
    boundary_nodes,
    exterior_boundaries,
    exterior_boundaries_as_a_set,
    flips,
    interior_boundaries,
    perimeter,
)
from .county_splits import CountySplit, county_splits
from .cut_edges import cut_edges, cut_edges_by_part
from .election import Election
from .flows import compute_edge_flows, flows_from_changes
from .tally import DataTally, Tally
from .spanning_trees import num_spanning_trees

__all__ = [
    "flows_from_changes",
    "county_splits",
    "cut_edges",
    "cut_edges_by_part",
    "Tally",
    "DataTally",
    "boundary_nodes",
    "flips",
    "perimeter",
    "exterior_boundaries",
    "interior_boundaries",
    "exterior_boundaries_as_a_set",
    "CountySplit",
    "compute_edge_flows",
    "Election",
    "num_spanning_trees"
]



import warnings

from shapely.strtree import STRtree


def neighbors(df, adjacency):
    if adjacency not in ("rook", "queen"):
        raise ValueError(
            "The adjacency parameter provided is not supported. "
            'We support "queen" or "rook" adjacency.'
        )

    return adjacencies[adjacency](df.geometry)


def str_tree(geometries):
    """Add ids to geometries and create a STR tree for spatial indexing.
    Use this for all spatial operations!
    """
    for i in geometries.index:
        geometries[i].id = i
    try:
        tree = STRtree(geometries)
    except AttributeError:
        tree = STRtree(geometries)
    return tree


def neighboring_geometries(geometries, tree=None):
    """Generator yielding tuples of the form (id, (ids of neighbors)).
    """
    if tree is None:
        tree = str_tree(geometries)

    for geometry in geometries:
        possible = tree.query(geometry)
        actual = tuple(
            p.id for p in possible if (not p.is_empty) and p.id != geometry.id
        )
        yield (geometry.id, actual)


def intersections_with_neighbors(geometries):
    """Generator yielding tuples of the form (id, {neighbor_id: intersection}).
    The intersections may be empty!
    """
    for i, neighbors in neighboring_geometries(geometries):
        intersections = {
            j: geometries[i].intersection(geometries[j]) for j in neighbors
        }
        yield (i, intersections)


def warn_for_overlaps(intersection_pairs):
    overlaps = set()
    for i, intersections in intersection_pairs:
        overlaps.update(
            set(
                tuple(sorted([i, j]))
                for j, intersection in intersections.items()
                if intersection.area > 0
            )
        )
        yield (i, intersections)
    if len(overlaps) > 0:
        warnings.warn(
            "Found overlaps among the given polygons. Indices of overlaps: {}".format(
                overlaps
            )
        )


def queen(geometries):
    """Return queen adjacency dictionary for the given collection of polygons."""
    intersection_pairs = warn_for_overlaps(intersections_with_neighbors(geometries))

    return {
        i: {
            j: {"shared_perim": intersection.length}
            for j, intersection in intersections.items()
            if (not intersection.is_empty)
        }
        for i, intersections in intersection_pairs
    }


def rook(geometries):
    """Return rook adjacency dictionary for the given collection of polygons."""
    return {
        i: {j: data for j, data in neighbors.items() if data["shared_perim"] > 0}
        for i, neighbors in queen(geometries).items()
    }


adjacencies = {"rook": rook, "queen": queen}



class Bounds:
    """
    Wrapper for numeric-validators to enforce upper and lower limits.

    This class is meant to be called as a function after instantiation; its
    return is ``True`` if the numeric validator is within set limits, and
    ``False`` otherwise.

    """

    def __init__(self, func, bounds):
        """
        :param func: Numeric validator function. Should return an iterable of values.
        :param bounds: Tuple of (lower, upper) numeric bounds.
        """
        self.func = func
        self.bounds = bounds

    def __call__(self, *args, **kwargs):
        lower, upper = self.bounds
        values = self.func(*args, **kwargs)
        return lower <= min(values) and max(values) <= upper

    @property
    def __name__(self):
        return "Bounds({},{})".format(self.func.__name__, str(self.bounds))


class UpperBound:
    """
    Wrapper for numeric-validators to enforce upper limits.

    This class is meant to be called as a function after instantiation; its
    return is ``True`` if the numeric validator is within a set upper limit,
    and ``False`` otherwise.

    """

    def __init__(self, func, bound):
        """
        :param func: Numeric validator function. Should return a comparable value.
        :param bounds: Comparable upper bound.
        """
        self.func = func
        self.bound = bound

    def __call__(self, *args, **kwargs):
        return self.func(*args, **kwargs) <= self.bound

    @property
    def __name__(self):
        return "UpperBound({} >= {})".format(self.func.__name__, self.bound)

    def __repr__(self):
        return "<{}>".format(self.__name__)


class LowerBound:
    """
    Wrapper for numeric-validators to enforce lower limits.

    This class is meant to be called as a function after instantiation; its
    return is ``True`` if the numeric validator is within a set lower limit,
    and ``False`` otherwise.

    """

    def __init__(self, func, bound):
        """
        :param func: Numeric validator function. Should return a comparable value.
        :param bounds: Comparable lower bound.
        """
        self.func = func
        self.bound = bound

    def __call__(self, *args, **kwargs):
        return self.func(*args, **kwargs) >= self.bound

    @property
    def __name__(self):
        return "LowerBound({} <= {})".format(self.func.__name__, self.bound)

    def __repr__(self):
        return "<{}>".format(self.__name__)


class SelfConfiguringUpperBound:
    """
    Wrapper for numeric-validators to enforce automatic upper limits.

    When instantiated, the initial upper bound is set as the initial value of
    the numeric-validator.

    This class is meant to be called as a function after instantiation; its
    return is ``True`` if the numeric validator is within a set upper limit,
    and ``False`` otherwise.

    """

    def __init__(self, func):
        """
        :param func: Numeric validator function.
        """
        self.func = func
        self.bound = None

    def __call__(self, partition):
        if not self.bound:
            self.bound = self.func(partition)
        return self.func(partition) <= self.bound


class SelfConfiguringLowerBound:
    """
    Wrapper for numeric-validators to enforce automatic lower limits.

    When instantiated, the initial lower bound is set as the initial value of
    the numeric-validator minus some configurable ε.

    This class is meant to be called as a function after instantiation; its
    return is ``True`` if the numeric validator is within a set lower limit,
    and ``False`` otherwise.

    """

    def __init__(self, func, epsilon=0.05):
        """
        :param func: Numeric validator function.
        :param epsilon: Initial "wiggle room" that the validator allows.
        """
        self.func = func
        self.bound = None
        self.epsilon = epsilon
        self.__name__ = func.__name__

    def __call__(self, partition):
        if not self.bound:
            self.bound = self.func(partition) - self.epsilon
        return self.func(partition) >= self.bound


class WithinPercentRangeOfBounds:
    def __init__(self, func, percent):
        self.func = func
        self.percent = float(percent) / 100.0
        self.lbound = None
        self.ubound = None

    def __call__(self, partition):
        if not (self.lbound and self.ubound):
            self.lbound = self.func(partition) * (1.0 - self.percent)
            self.ubound = self.func(partition) * (1.0 + self.percent)
            return True
        else:
            return self.lbound <= self.func(partition) <= self.ubound



import functools
import json
from typing import Any
import warnings

import geopandas as gp
import networkx
from networkx.classes.function import frozen
from networkx.readwrite import json_graph
from shapely.ops import unary_union
from shapely.prepared import prep

from .adjacency import neighbors
from .geo import GeometryError, invalid_geometries, reprojected


class Graph(networkx.Graph):
    """Represents a graph to be partitioned. It is based on :class:`networkx.Graph`.

    We have added some classmethods to help construct graphs from shapefiles, and
    to save and load graphs as JSON files.

    """
    def __repr__(self):
        return "<Graph [{} nodes, {} edges]>".format(len(self.nodes), len(self.edges))

    @classmethod
    def from_networkx(cls, graph: networkx.Graph):
        g = cls(graph)
        return g

    @classmethod
    def from_json(cls, json_file):
        """Load a graph from a JSON file in the NetworkX json_graph format.
        :param json_file: Path to JSON file.
        :return: Graph
        """
        with open(json_file) as f:
            data = json.load(f)
        g = json_graph.adjacency_graph(data)
        graph = cls.from_networkx(g)
        graph.issue_warnings()
        return graph

    def to_json(self, json_file, *, include_geometries_as_geojson=False):
        """Save a graph to a JSON file in the NetworkX json_graph format.

        :param json_file: Path to target JSON file.
        :param bool include_geometry_as_geojson: (optional) Whether to include
            any :mod:`shapely` geometry objects encountered in the graph's node
            attributes as GeoJSON. The default (``False``) behavior is to remove
            all geometry objects because they are not serializable. Including the
            GeoJSON will result in a much larger JSON file.
        """
        data = json_graph.adjacency_data(self)

        if include_geometries_as_geojson:
            convert_geometries_to_geojson(data)
        else:
            remove_geometries(data)

        with open(json_file, "w") as f:
            json.dump(data, f)

    @classmethod
    def from_file(
        cls,
        filename,
        adjacency="rook",
        cols_to_add=None,
        reproject=False,
        ignore_errors=False,
    ):
        """Create a :class:`Graph` from a shapefile (or GeoPackage, or GeoJSON, or
        any other library that :mod:`geopandas` can read. See :meth:`from_geodataframe`
        for more details.

        :param cols_to_add: (optional) The names of the columns that you want to
            add to the graph as node attributes. By default, all columns are added.
        """
        df = gp.read_file(filename)
        graph = cls.from_geodataframe(
            df, adjacency=adjacency,
            cols_to_add=cols_to_add,
            reproject=reproject,
            ignore_errors=ignore_errors
        )
        graph.graph["crs"] = df.crs.to_json()
        return graph

    @classmethod
    def from_geodataframe(
        cls,
        dataframe,
        adjacency="rook",
        cols_to_add=None,
        reproject=False,
        ignore_errors=False
    ):
        """Creates the adjacency :class:`Graph` of geometries described by `dataframe`.
        The areas of the polygons are included as node attributes (with key `area`).
        The shared perimeter of neighboring polygons are included as edge attributes
        (with key `shared_perim`).
        Nodes corresponding to polygons on the boundary of the union of all the geometries
        (e.g., the state, if your dataframe describes VTDs) have a `boundary_node` attribute
        (set to `True`) and a `boundary_perim` attribute with the length of this "exterior"
        boundary.

        By default, areas and lengths are computed in a UTM projection suitable for the
        geometries. This prevents the bizarro area and perimeter values that show up when
        you accidentally do computations in Longitude-Latitude coordinates. If the user
        specifies `reproject=False`, then the areas and lengths will be computed in the
        GeoDataFrame's current coordinate reference system. This option is for users who
        have a preferred CRS they would like to use.

        :param dataframe: :class:`geopandas.GeoDataFrame`
        :param adjacency: (optional) The adjacency type to use ("rook" or "queen").
            Default is "rook"
        :param cols_to_add: (optional) The names of the columns that you want to
            add to the graph as node attributes. By default, all columns are added.
        :param reproject: (optional) Whether to reproject to a UTM projection before
            creating the graph. Default is ``True``.
        :param ignore_errors: (optional) Whether to ignore all invalid geometries and
            attept to create the graph anyway. Default is ``False``.
        :return: The adjacency graph of the geometries from `dataframe`.
        :rtype: :class:`Graph`
        """
        # Validate geometries before reprojection
        if not ignore_errors:
            invalid = invalid_geometries(dataframe)
            if len(invalid) > 0:
                raise GeometryError(
                    "Invalid geometries at rows {} before "
                    "reprojection. Consider repairing the affected geometries with "
                    "`.buffer(0)`, or pass `ignore_errors=True` to attempt to create "
                    "the graph anyways.".format(invalid)
                )

        # Project the dataframe to an appropriate UTM projection unless
        # explicitly told not to.
        if reproject:
            df = reprojected(dataframe)
            if ignore_errors:
                invalid_reproj = invalid_geometries(df)
                print(invalid_reproj)
                if len(invalid_reproj) > 0:
                    raise GeometryError(
                        "Invalid geometries at rows {} after "
                        "reprojection. Consider reloading the GeoDataFrame with "
                        "`reproject=False` or repairing the affected geometries "
                        "with `.buffer(0)`.".format(invalid_reproj)
                    )
        else:
            df = dataframe

        # Generate dict of dicts of dicts with shared perimeters according
        # to the requested adjacency rule
        adjacencies = neighbors(df, adjacency)
        graph = cls(adjacencies)

        graph.geometry = df.geometry

        graph.issue_warnings()

        # Add "exterior" perimeters to the boundary nodes
        add_boundary_perimeters(graph, df.geometry)

        # Add area data to the nodes
        areas = df.geometry.area.to_dict()
        networkx.set_node_attributes(graph, name="area", values=areas)

        graph.add_data(df, columns=cols_to_add)
        graph.graph["crs"] = df.crs.to_json()
        return graph

    def lookup(self, node, field):
        """
        Lookup a node/field attribute.
        :param node: Node to look up.
        :param field: Field to look up.
        """
        return self.nodes[node][field]

    @property
    def node_indices(self):
        return set(self.nodes)

    @property
    def edge_indices(self):
        return set(self.edges)

    def add_data(self, df, columns=None):
        """Add columns of a DataFrame to a graph as node attributes using
        by matching the DataFrame's index to node ids.

        :param df: Dataframe containing given columns.
        :param columns: (optional) List of dataframe column names to add.
        """

        if columns is None:
            columns = list(df.columns)

        check_dataframe(df[columns])

        column_dictionaries = df.to_dict("index")
        networkx.set_node_attributes(self, column_dictionaries)

        if hasattr(self, "data"):
            self.data[columns] = df[columns]
        else:
            self.data = df[columns]

    def join(self, dataframe, columns=None, left_index=None, right_index=None):
        """Add data from a dataframe to the graph, matching nodes to rows when
        the node's `left_index` attribute equals the row's `right_index` value.

        :param dataframe: DataFrame.
        :columns: (optional) The columns whose data you wish to add to the graph.
            If not provided, all columns are added.
        :left_index: (optional) The node attribute used to match nodes to rows.
            If not provided, node IDs are used.
        :right_index: (optional) The DataFrame column name to use to match rows
            to nodes. If not provided, the DataFrame's index is used.
        """
        if right_index is not None:
            df = dataframe.set_index(right_index)
        else:
            df = dataframe

        if columns is not None:
            df = df[columns]

        check_dataframe(df)

        column_dictionaries = df.to_dict()

        if left_index is not None:
            ids_to_index = networkx.get_node_attributes(self, left_index)
        else:
            # When the left_index is node ID, the matching is just
            # a redundant {node: node} dictionary
            ids_to_index = dict(zip(self.nodes, self.nodes))

        node_attributes = {
            node_id: {
                column: values[index] for column, values in column_dictionaries.items()
            }
            for node_id, index in ids_to_index.items()
        }

        networkx.set_node_attributes(self, node_attributes)

    @property
    def islands(self):
        """The set of degree-0 nodes."""
        return set(node for node in self if self.degree[node] == 0)

    def warn_for_islands(self):
        """Issue a warning if the graph has any islands (degree-0 nodes)."""
        islands = self.islands
        if len(self.islands) > 0:
            warnings.warn(
                "Found islands (degree-0 nodes). Indices of islands: {}".format(islands)
            )

    def issue_warnings(self):
        """Issue warnings if the graph has any red flags (right now, only islands)."""
        self.warn_for_islands()


def add_boundary_perimeters(graph, geometries):
    """Add shared perimeter between nodes and the total geometry boundary.

    :param graph: NetworkX graph
    :param df: Geodataframe containing geometry information.
    :return: The updated graph.
    """
    prepared_boundary = prep(unary_union(geometries).boundary)

    boundary_nodes = geometries.boundary.apply(prepared_boundary.intersects)

    for node in graph:
        graph.nodes[node]["boundary_node"] = bool(boundary_nodes[node])
        if boundary_nodes[node]:
            total_perimeter = geometries[node].boundary.length
            shared_perimeter = sum(
                neighbor_data["shared_perim"] for neighbor_data in graph[node].values()
            )
            boundary_perimeter = total_perimeter - shared_perimeter
            graph.nodes[node]["boundary_perim"] = boundary_perimeter


def check_dataframe(df):
    for column in df.columns:
        if sum(df[column].isna()) > 0:
            warnings.warn("NA values found in column {}!".format(column))


def remove_geometries(data):
    """Remove geometry attributes from NetworkX adjacency data object,
    because they are not serializable. Mutates the ``data`` object.

    Does nothing if no geometry attributes are found.

    :param data: an adjacency data object (returned by
        :func:`networkx.readwrite.json_graph.adjacency_data`)
    """
    for node in data["nodes"]:
        bad_keys = []
        for key in node:
            # having a ``__geo_interface__``` property identifies the object
            # as being a ``shapely`` geometry object
            if hasattr(node[key], "__geo_interface__"):
                bad_keys.append(key)
        for key in bad_keys:
            del node[key]


def convert_geometries_to_geojson(data):
    """Convert geometry attributes in a NetworkX adjacency data object
    to GeoJSON, so that they can be serialized. Mutates the ``data`` object.

    Does nothing if no geometry attributes are found.

    :param data: an adjacency data object (returned by
        :func:`networkx.readwrite.json_graph.adjacency_data`)
    """
    for node in data["nodes"]:
        for key in node:
            # having a ``__geo_interface__``` property identifies the object
            # as being a ``shapely`` geometry object
            if hasattr(node[key], "__geo_interface__"):
                # The ``__geo_interface__`` property is essentially GeoJSON.
                # This is what :func:`geopandas.GeoSeries.to_json` uses under
                # the hood.
                node[key] = node[key].__geo_interface__


class FrozenGraph:
    """ Represents an immutable graph to be partitioned. It is based off :class:`Graph`.

    This speeds up chain runs and prevents having to deal with cache invalidation issues.
    This class behaves slightly differently than :class:`Graph` or :class:`networkx.Graph`.
    """
    __slots__ = [
        "graph",
        "size"
    ]

    def __init__(self, graph: Graph):
        self.graph = networkx.classes.function.freeze(graph)
        self.graph.join = frozen
        self.graph.add_data = frozen

        self.size = len(self.graph)

    def __len__(self):
        return self.size

    def __getattribute__(self, __name: str) -> Any:
        try:
            return object.__getattribute__(self, __name)
        except AttributeError:
            return object.__getattribute__(self.graph, __name)

    def __getitem__(self, __name: str) -> Any:
        return self.graph[__name]

    def __iter__(self):
        yield from self.node_indices

    @functools.lru_cache(16384)
    def neighbors(self, n):
        return tuple(self.graph.neighbors(n))

    @functools.cached_property
    def node_indices(self):
        return self.graph.node_indices

    @functools.cached_property
    def edge_indices(self):
        return self.graph.edge_indices

    @functools.lru_cache(16384)
    def degree(self, n):
        return self.graph.degree(n)

    @functools.lru_cache(65536)
    def lookup(self, node, field):
        return self.graph.nodes[node][field]

    def subgraph(self, nodes):
        return FrozenGraph(self.graph.subgraph(nodes))



import numpy


def mean_median(election_results):
    """
    Computes the Mean-Median score for the given ElectionResults.
    A positive value indicates an advantage for the first party listed
    in the Election's parties_to_columns dictionary.
    """
    first_party = election_results.election.parties[0]
    data = election_results.percents(first_party)

    return numpy.median(data) - numpy.mean(data)


def mean_thirdian(election_results):
    """
    Computes the Mean-Median score for the given ElectionResults.
    A positive value indicates an advantage for the first party listed
    in the Election's parties_to_columns dictionary.

    The motivation for this score is that the minority party in many
    states struggles to win even a third of the seats.
    """
    first_party = election_results.election.parties[0]
    data = election_results.percents(first_party)

    thirdian_index = round(len(data) / 3)
    thirdian = sorted(data)[thirdian_index]

    return thirdian - numpy.mean(data)


def efficiency_gap(results):
    """
    Computes the efficiency gap for the given ElectionResults.
    A positive value indicates an advantage for the first party listed
    in the Election's parties_to_columns dictionary.
    """
    party1, party2 = [results.counts(party) for party in results.election.parties]
    wasted_votes_by_part = map(wasted_votes, party1, party2)
    total_votes = results.total_votes()
    numerator = sum(waste2 - waste1 for waste1, waste2 in wasted_votes_by_part)
    return numerator / total_votes


def wasted_votes(party1_votes, party2_votes):
    """
    Computes the wasted votes for each party in the given race.
    :party1_votes: the number of votes party1 received in the race
    :party2_votes: the number of votes party2 received in the race
    """
    total_votes = party1_votes + party2_votes
    if party1_votes > party2_votes:
        party1_waste = party1_votes - total_votes / 2
        party2_waste = party2_votes
    else:
        party2_waste = party2_votes - total_votes / 2
        party1_waste = party1_votes
    return party1_waste, party2_waste


def partisan_bias(election_results):
    """
    Computes the partisan bias for the given ElectionResults.
    The partisan bias is defined as the number of districts with above-mean
    vote share by the first party divided by the total number of districts,
    minus 1/2.
    """
    first_party = election_results.election.parties[0]
    party_shares = numpy.array(election_results.percents(first_party))
    mean_share = numpy.mean(party_shares)
    above_mean_districts = len(party_shares[party_shares > mean_share])
    return (above_mean_districts / len(party_shares)) - 0.5


def partisan_gini(election_results):
    """
    Computes the partisan Gini score for the given ElectionResults.
    The partisan Gini score is defined as the area between the seats-votes
    curve and its reflection about (.5, .5).
    """
    # For two parties, the Gini score is symmetric--it does not vary by party.
    party = election_results.election.parties[0]

    # To find seats as a function of votes, we assume uniform partisan swing.
    # That is, if the statewide popular vote share for a party swings by some
    # delta, the vote share for that party swings by that delta in each
    # district.
    # We calculate the necessary delta to shift the district with the highest
    # vote share for the party to a vote share of 0.5. This delta, subtracted
    # from the original popular vote share, gives the minimum popular vote
    # share that yields 1 seat to the party.
    # We repeat this process for the district with the second-highest vote
    # share, which gives the minimum popular vote share yielding 2 seats,
    # and so on.
    overall_result = election_results.percent(party)
    race_results = sorted(election_results.percents(party), reverse=True)
    seats_votes = [overall_result - r + 0.5 for r in race_results]

    # Apply reflection of seats-votes curve about (.5, .5)
    reflected_sv = reversed([1 - s for s in seats_votes])
    # Calculate the unscaled, unsigned area between the seats-votes curve
    # and its reflection. For each possible number of seats attained, we find
    # the area of a rectangle of unit height, with a width determined by the
    # horizontal distance between the curves at that number of seats.
    unscaled_area = sum(abs(s - r) for s, r in zip(seats_votes, reflected_sv))
    # We divide by area by the number of seats to obtain a partisan Gini score
    # between 0 and 1.
    return unscaled_area / len(race_results)



import collections

from .flows import on_edge_flow, neighbor_flips


def put_edges_into_parts(edges, assignment):
    by_part = collections.defaultdict(set)
    for edge in edges:
        # add edge to the sets corresponding to the parts it touches
        by_part[assignment.mapping[edge[0]]].add(edge)
        by_part[assignment.mapping[edge[1]]].add(edge)
    return by_part


def new_cuts(partition):
    """The edges that were not cut, but now are"""
    return {
        (node, neighbor)
        for node, neighbor in neighbor_flips(partition)
        if partition.crosses_parts((node, neighbor))
    }


def obsolete_cuts(partition):
    """The edges that were cut, but now are not"""
    return {
        (node, neighbor)
        for node, neighbor in neighbor_flips(partition)
        if partition.parent.crosses_parts((node, neighbor))
        and not partition.crosses_parts((node, neighbor))
    }


def initialize_cut_edges(partition):
    edges = {
        tuple(sorted(edge))
        for edge in partition.graph.edges
        if partition.crosses_parts(edge)
    }
    return put_edges_into_parts(edges, partition.assignment)


@on_edge_flow(initialize_cut_edges, alias="cut_edges_by_part")
def cut_edges_by_part(partition, previous, new_edges, old_edges):
    return (previous | new_edges) - old_edges


def cut_edges(partition):
    parent = partition.parent

    if not parent:
        return {
            tuple(sorted(edge))
            for edge in partition.graph.edges
            if partition.crosses_parts(edge)
        }
    # Edges that weren't cut, but now are cut
    # We sort the tuples to make sure we don't accidentally end
    # up with both (4,5) and (5,4) (for example) in it
    new, obsolete = new_cuts(partition), obsolete_cuts(partition)

    return (parent["cut_edges"] | new) - obsolete



from collections import Counter
from shapely.validation import explain_validity
from gerrychain.vendor.utm import from_latlon


def utm_of_point(point):
    return from_latlon(point.y, point.x)[2]


def identify_utm_zone(df):
    wgs_df = df.to_crs("+proj=longlat +ellps=WGS84 +datum=WGS84 +no_defs")
    utm_counts = Counter(utm_of_point(point) for point in wgs_df["geometry"].centroid)
    # most_common returns a list of tuples, and we want the 0,0th entry
    most_common = utm_counts.most_common(1)[0][0]
    return most_common


def invalid_geometries(df):
    """Given a GeoDataFrame, returns a list of row indices
    with invalid geometries.

    :param df: :class:`geopandas.GeoDataFrame`
    :rtype: list of int
    """
    invalid = []
    for idx, row in df.iterrows():
        validity = explain_validity(row.geometry)
        if validity != "Valid Geometry":
            invalid.append(idx)
    return invalid


def reprojected(df):
    """Returns a copy of `df`, projected into the coordinate reference system of a suitable
        `Universal Transverse Mercator`_ zone.
    :param df: :class:`geopandas.GeoDataFrame`
    :rtype: :class:`geopandas.GeoDataFrame`

    .. _`Universal Transverse Mercator`: https://en.wikipedia.org/wiki/UTM_coordinate_system
    """
    utm = identify_utm_zone(df)
    return df.to_crs(
        "+proj=utm +zone={utm} +ellps=WGS84 +datum=WGS84 +units=m +no_defs".format(
            utm=utm
        )
    )


class GeometryError(Exception):
    """
    Wrapper error class for projection failures.
    Changing a map's projection may create invalid geometries, which may
    or may not be repairable using the `.buffer(0)`_ trick.

    .. _`.buffer(0)`: https://shapely.readthedocs.io/en/stable/manual.html#constructive-methods
    """



import collections
import math
import warnings

from .flows import flows_from_changes, on_flow


class DataTally:
    """An updater for tallying numerical data that is not necessarily stored as
    node attributes
    """
    __slots__ = [
        "data",
        "alias",
        "_call"
    ]

    def __init__(self, data, alias):
        """
        :param data: a dict or Series indexed by the graph's nodes,
            or the string key for a node attribute containing the Tally's data.
        :param alias: the name of the tally in the Partition's `updaters` dictionary
        """
        self.data = data
        self.alias = alias

        def initialize_tally(partition):
            if isinstance(self.data, str):
                nodes = partition.graph.nodes
                attribute = self.data
                self.data = {node: nodes[node][attribute] for node in nodes}

            tally = collections.defaultdict(int)
            for node, part in partition.assignment.items():
                add = self.data[node]

                if math.isnan(add):
                    warnings.warn(
                        "ignoring nan encountered at node '{}' for attribute '{}'".format(
                            node, self.alias
                        )
                    )
                else:
                    tally[part] += add
            return dict(tally)

        @on_flow(initialize_tally, alias=alias)
        def update_tally(partition, previous, new_nodes, old_nodes):
            inflow = sum(self.data[node] for node in new_nodes)
            outflow = sum(self.data[node] for node in old_nodes)
            return previous + inflow - outflow

        self._call = update_tally

    def __call__(self, partition, previous=None):
        return self._call(partition, previous)


class Tally:
    """An updater for keeping a tally of one or more node attributes.
    """
    __slots__ = [
        "fields",
        "alias",
        "dtype"
    ]

    def __init__(self, fields, alias=None, dtype=int):
        """
        :param fields: the list of node attributes that you want to tally. Or a just a
            single attribute name as a string.
        :param alias: the aliased name of this Tally (meaning, the key corresponding to
            this Tally in the Partition's updaters dictionary)
        :param dtype: the type (int, float, etc.) that you want the tally to have
        """
        if not isinstance(fields, list):
            fields = [fields]
        if not alias:
            alias = fields[0]
        self.fields = fields
        self.alias = alias
        self.dtype = dtype

    def __call__(self, partition):
        if partition.parent is None:
            return self._initialize_tally(partition)
        return self._update_tally(partition)

    def _initialize_tally(self, partition):
        """Compute the initial district-wide tally of data stored in the "field"
        attribute of nodes.

        :param partition: :class:`Partition` class.
        """
        tally = collections.defaultdict(self.dtype)
        for node, part in partition.assignment.items():
            add = self._get_tally_from_node(partition, node)

            if math.isnan(add):
                warnings.warn(
                    "ignoring nan encountered at node '{}' for attribute '{}' "
                    "with fields {}".format(node, self.alias, self.fields)
                )
            else:
                tally[part] += add
        return dict(tally)

    def _update_tally(self, partition):
        """Compute the district-wide tally of data stored in the "field" attribute
        of nodes, given proposed changes.

        :param partition: :class:`Partition` class.
        """
        parent = partition.parent

        old_tally = parent[self.alias]
        new_tally = dict(old_tally)

        graph = partition.graph

        for part, flow in flows_from_changes(parent, partition).items():
            out_flow = compute_out_flow(graph, self.fields, flow)
            in_flow = compute_in_flow(graph, self.fields, flow)
            new_tally[part] = old_tally[part] - out_flow + in_flow

        return new_tally

    def _get_tally_from_node(self, partition, node):
        return sum(partition.graph.lookup(node, field) for field in self.fields)


def compute_out_flow(graph, fields, flow):
    return sum(graph.lookup(node, field) for node in flow["out"] for field in fields)


def compute_in_flow(graph, fields, flow):
    return sum(graph.lookup(node, field) for node in flow["in"] for field in fields)




import math

from .bounds import SelfConfiguringLowerBound, SelfConfiguringUpperBound


def L1_reciprocal_polsby_popper(partition):
    return sum(1 / value for value in partition["polsby_popper"].values())


def L1_polsby_popper(partition):
    return sum(value for value in partition["polsby_popper"].values())


def L2_polsby_popper(partition):
    return math.sqrt(sum(value ** 2 for value in partition["polsby_popper"].values()))


def L_minus_1_polsby_popper(partition):
    return len(partition.parts) / sum(
        1 / value for value in partition["polsby_popper"].values()
    )


no_worse_L_minus_1_polsby_popper = SelfConfiguringLowerBound(L_minus_1_polsby_popper)

no_worse_L1_reciprocal_polsby_popper = SelfConfiguringUpperBound(
    L1_reciprocal_polsby_popper
)



from setuptools import find_packages, setup

import versioneer

with open("./README.rst") as f:
    long_description = f.read()

requirements = [
    # package requirements go here
    "pandas",
    "scipy",
    "networkx",
    "geopandas",
    "shapely",
    "matplotlib",
]

setup(
    name="gerrychain",
    description="Use Markov chain Monte Carlo to analyze districting plans and gerrymanders",
    author="Metric Geometry and Gerrymandering Group",
    author_email="gerrymandr@gmail.com",
    maintainer="Max Hully",
    maintainer_email="max@mggg.org",
    long_description=long_description,
    long_description_content_type="text/x-rst",
    url="https://github.com/mggg/GerryChain",
    packages=find_packages(exclude=("tests",)),
    version=versioneer.get_version(),
    cmdclass=versioneer.get_cmdclass(),
    install_requires=requirements,
    keywords="GerryChain",
    classifiers=[
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Operating System :: OS Independent",
        "License :: OSI Approved :: BSD License",
    ],
)



import math


def compute_polsby_popper(area, perimeter):
    try:
        return 4 * math.pi * area / perimeter ** 2
    except ZeroDivisionError:
        return math.nan


def polsby_popper(partition):
    """Computes Polsby-Popper compactness scores for each district in the partition.
    """
    return {
        part: compute_polsby_popper(
            partition["area"][part], partition["perimeter"][part]
        )
        for part in partition.parts
    }



"""
The :mod:`gerrychain.constraints` module provides a collection of constraint
functions and helper classes for the validation step in GerryChain.

==================================== ==========================================
Helper classes
===============================================================================
:class:`Validator`                    Collection of constraints
:class:`Bounds`                       Bounds on numeric constraints
:class:`UpperBound`                  Upper bounds on numeric constraints
:class:`LowerBound`                  Lower bounds on numeric constraints
:class:`SelfConfiguringUpperBound`    Automatic upper bounds on numeric constraints
:class:`SelfConfiguringLowerBound`    Automatic lower bounds on numeric constraints
:class:`WithinPercentRangeOfBounds`   Percentage bounds for numeric constraints
==================================== ==========================================

|

=================================================== =============================================
Binary constraint functions
=================================================================================================
:meth:`no_worse_L1_reciprocal_polsby_popper`            Lower bounded L1-reciprocal Polsby-Popper
:meth:`no_worse_L_minus_1_reciprocal_polsby_popper`     Lower bounded L(-1)-reciprocal Polsby-Popper
:meth:`contiguous`                                      Districts are contiguous (with NetworkX methods)
:meth:`single_flip_contiguous`                          Districts are contiguous (optimized for ``propose_random_flip`` proposal)
:meth:`no_vanishing_districts`                          No districts may be completely consumed
=================================================== =============================================

Each new step proposed to the chain is passed off to the "validator" functions
here to determine whether or not the step is valid. If it is invalid (breaks
contiguity, for instance), then the step is immediately rejected.

A validator should take in a :class:`~gerrychain.partition.Partition` instance,
and should return whether or not the instance is valid according to their
rules. Many top-level functions following this signature in this module are
examples of this.

"""

from .bounds import (LowerBound, SelfConfiguringLowerBound,
                     SelfConfiguringUpperBound, UpperBound,
                     WithinPercentRangeOfBounds, Bounds)
from .compactness import (L1_polsby_popper, L1_reciprocal_polsby_popper,
                          L2_polsby_popper, L_minus_1_polsby_popper,
                          no_worse_L1_reciprocal_polsby_popper,
                          no_worse_L_minus_1_polsby_popper)
from .contiguity import (contiguous, contiguous_bfs, no_more_discontiguous,
                         single_flip_contiguous)
from .validity import (Validator, districts_within_tolerance,
                       no_vanishing_districts, refuse_new_splits,
                       within_percent_of_ideal_population)

__all__ = ["LowerBound", "SelfConfiguringLowerBound",
        "SelfConfiguringUpperBound", "UpperBound",
        "WithinPercentRangeOfBounds", "L1_polsby_popper",
        "L1_reciprocal_polsby_popper", "L2_polsby_popper",
        "L_minus_1_polsby_popper", "no_worse_L1_reciprocal_polsby_popper",
        "no_worse_L_minus_1_polsby_popper", "contiguous", "contiguous_bfs",
        "no_more_discontiguous", "single_flip_contiguous", "Validator",
        "districts_within_tolerance", "no_vanishing_districts",
        "refuse_new_splits", "within_percent_of_ideal_population", "Bounds"]



from collections import defaultdict
from collections.abc import Mapping

import pandas


class Assignment(Mapping):
    """An assignment of nodes into parts.

    The goal of :class:`Assignment` is to provide an interface that mirrors a
    dictionary (what we have been using for assigning nodes to districts) while making it
    convenient/cheap to access the set of nodes in each part.

    An :class:`Assignment` has a ``parts`` property that is a dictionary of the form
    ``{part: <frozenset of nodes in part>}``.
    """
    __slots__ = [
        'parts',
        'mapping'
    ]

    def __init__(self, parts, mapping=None, validate=True):
        if validate:
            number_of_keys = sum(len(keys) for keys in parts.values())
            number_of_unique_keys = len(set().union(*parts.values()))
            if number_of_keys != number_of_unique_keys:
                raise ValueError("Keys must have unique assignments.")
            if not all(isinstance(keys, frozenset) for keys in parts.values()):
                raise TypeError("Level sets must be frozensets")
        self.parts = parts

        if not mapping:
            self.mapping = {}
            for part, nodes in self.parts.items():
                for node in nodes:
                    self.mapping[node] = part
        else:
            self.mapping = mapping

    def __repr__(self):
        return "<Assignment [{} keys, {} parts]>".format(len(self), len(self.parts))

    def __iter__(self):
        return self.keys()

    def __len__(self):
        return sum(len(keys) for keys in self.parts.values())

    def __getitem__(self, node):
        return self.mapping[node]

    def copy(self):
        """Returns a copy of the assignment.
        Does not duplicate the frozensets of nodes, just the parts dictionary.
        """
        return Assignment(self.parts.copy(), self.mapping.copy(), validate=False)

    def update_flows(self, flows):
        """Update the assignment for some nodes using the given flows.
        """
        for part, flow in flows.items():
            # Union between frozenset and set returns an object whose type
            # matches the object on the left, which here is a frozenset
            self.parts[part] = (self.parts[part] - flow["out"]) | flow["in"]

            for node in flow["in"]:
                self.mapping[node] = part

    def items(self):
        """Iterate over ``(node, part)`` tuples, where ``node`` is assigned to ``part``.
        """
        yield from self.mapping.items()

    def keys(self):
        yield from self.mapping.keys()

    def values(self):
        yield from self.mapping.values()

    def update_parts(self, new_parts):
        """Update some parts of the assignment. Does not check that every node is
        still assigned to a part.

        :param dict new_parts: dictionary mapping (some) parts to their new sets or
            frozensets of nodes
        """
        for part, nodes in new_parts.items():
            self.parts[part] = frozenset(nodes)

            for node in nodes:
                self.mapping[node] = part

    def to_series(self):
        """Convert the assignment to a :class:`pandas.Series`."""
        groups = [
            pandas.Series(data=part, index=nodes) for part, nodes in self.parts.items()
        ]
        return pandas.concat(groups)

    def to_dict(self):
        """Convert the assignment to a ``{node: part}`` dictionary."""
        return self.mapping

    @classmethod
    def from_dict(cls, assignment):
        """Create an :class:`Assignment` from a dictionary. This is probably the method you want
        to use to create a new assignment.

        This also works for :class:`pandas.Series`.
        """
        parts = {part: frozenset(keys) for part, keys in level_sets(assignment).items()}

        return cls(parts)


def get_assignment(assignment, graph=None):
    if isinstance(assignment, str):
        if graph is None:
            raise TypeError(
                "You must provide a graph when using a node attribute for the assignment"
            )
        return Assignment.from_dict(
            {node: graph.nodes[node][assignment] for node in graph}
        )
    elif callable(getattr(assignment, "items", None)):
        return Assignment.from_dict(assignment)
    elif isinstance(assignment, Assignment):
        return assignment
    else:
        raise TypeError("Assignment must be a dict or a node attribute key")


def level_sets(mapping: dict, container=set):
    """Inverts a dictionary. ``{key: value}`` becomes
    ``{value: <container of keys that map to value>}``."""
    sets = defaultdict(container)
    for source, target in mapping.items():
        sets[target].add(source)
    return sets



from functools import partial
from ..random import random

from ..tree import (
    recursive_tree_part, bipartition_tree, bipartition_tree_random,
    _bipartition_tree_random_all, uniform_spanning_tree,
    find_balanced_edge_cuts_memoization,
)


def recom(
    partition, pop_col, pop_target, epsilon, node_repeats=1, method=bipartition_tree
):
    """ReCom proposal.

    Description from MGGG's 2018 Virginia House of Delegates report:
    At each step, we (uniformly) randomly select a pair of adjacent districts and
    merge all of their blocks in to a single unit. Then, we generate a spanning tree
    for the blocks of the merged unit with the Kruskal/Karger algorithm. Finally,
    we cut an edge of the tree at random, checking that this separates the region
    into two new districts that are population balanced.

    Example usage::

        from functools import partial
        from gerrychain import MarkovChain
        from gerrychain.proposals import recom

        # ...define constraints, accept, partition, total_steps here...

        # Ideal population:
        pop_target = sum(partition["population"].values()) / len(partition)

        proposal = partial(
            recom, pop_col="POP10", pop_target=pop_target, epsilon=.05, node_repeats=10
        )

        chain = MarkovChain(proposal, constraints, accept, partition, total_steps)

    """
    edge = random.choice(tuple(partition["cut_edges"]))
    parts_to_merge = (partition.assignment.mapping[edge[0]], partition.assignment.mapping[edge[1]])

    subgraph = partition.graph.subgraph(
        partition.parts[parts_to_merge[0]] | partition.parts[parts_to_merge[1]]
    )

    flips = recursive_tree_part(
        subgraph.graph,
        parts_to_merge,
        pop_col=pop_col,
        pop_target=pop_target,
        epsilon=epsilon,
        node_repeats=node_repeats,
        method=method,
    )

    return partition.flip(flips)


def reversible_recom(partition, pop_col, pop_target, epsilon,
                     balance_edge_fn=find_balanced_edge_cuts_memoization, M=1,
                     repeat_until_valid=False, choice=random.choice):
    """Reversible ReCom proposal."""
    def dist_pair_edges(part, a, b):
        return set(
            e for e in part.graph.edges
            if ((part.assignment.mapping[e[0]] == a and part.assignment.mapping[e[1]] == b) or
                (part.assignment.mapping[e[0]] == b and part.assignment.mapping[e[1]] == a))
        )

    def bounded_balance_edge_fn(*args, **kwargs):
        cuts = balance_edge_fn(*args, **kwargs)
        if len(cuts) > M:
            raise ReversibilityError(f'Found {len(cuts)} balance edges, '
                                     f'but the upper bound is {M}.')
        return cuts

    bipartition_tree_random_reversible = partial(
        _bipartition_tree_random_all,
        repeat_until_valid=repeat_until_valid,
        spanning_tree_fn=uniform_spanning_tree,
        balance_edge_fn=bounded_balance_edge_fn
    )

    parts = sorted(list(partition.parts.keys()))
    dist_pairs = []
    for out_part in parts:
        for in_part in parts:
            dist_pairs.append((out_part, in_part))

    random_pair = random.choice(dist_pairs)
    pair_edges = dist_pair_edges(partition, *random_pair)
    if random_pair[0] == random_pair[1] or not pair_edges:
        return partition    # self-loop: no adjacency

    edge = random.choice(list(pair_edges))
    parts_to_merge = (partition.assignment.mapping[edge[0]], partition.assignment.mapping[edge[1]])
    subgraph = partition.graph.subgraph(
        partition.parts[parts_to_merge[0]] | partition.parts[parts_to_merge[1]]
    )

    all_cuts = bipartition_tree_random_reversible(
        subgraph,
        pop_col=pop_col,
        pop_target=pop_target,
        epsilon=epsilon
    )
    if not all_cuts:
        return partition    # self-loop: no balance edge

    nodes = choice(all_cuts).subset
    remaining_nodes = set(subgraph.nodes()) - set(nodes)
    flips = {
        **{node: parts_to_merge[0] for node in nodes},
        **{node: parts_to_merge[1] for node in remaining_nodes}
    }

    new_part = partition.flip(flips)
    seam_length = len(dist_pair_edges(new_part, *random_pair))

    prob = len(all_cuts) / (M * seam_length)
    if prob > 1:
        raise ReversibilityError(f'Found {len(all_cuts)} balance edges, but '
                                 f'the upper bound (with seam length 1) is {M}.')
    if random.random() < prob:
        return new_part

    return partition     # self-loop


class ReCom:
    def __init__(self, pop_col, ideal_pop, epsilon, method=bipartition_tree_random):
        self.pop_col = pop_col
        self.ideal_pop = ideal_pop
        self.epsilon = epsilon
        self.method = method

    def __call__(self, partition):
        return recom(
            partition, self.pop_col, self.ideal_pop, self.epsilon, method=self.method
        )


class ReversibilityError(Exception):
    """Raised when the cut edge upper bound is violated."""
    def __init__(self, msg):
        self.message = msg



class SubgraphView:
    __slots__ = [
        "graph",
        "parts",
        "subgraphs_cache"
    ]

    def __init__(self, graph, parts):
        self.graph = graph
        self.parts = parts
        self.subgraphs_cache = {}

    def __getitem__(self, part):
        if part not in self.subgraphs_cache:
            self.subgraphs_cache[part] = self.graph.subgraph(self.parts[part])
        return self.subgraphs_cache[part]

    def __iter__(self):
        for part in self.parts:
            yield self[part]

    def items(self):
        for part in self.parts:
            yield part, self[part]



from .compactness import polsby_popper
from .partisan import (
    mean_median,
    partisan_bias,
    partisan_gini,
    efficiency_gap,
    wasted_votes,
)

__all__ = [
    "mean_median",
    "partisan_bias",
    "partisan_gini",
    "efficiency_gap",
    "polsby_popper",
    "wasted_votes",
]

